
from argparse import ArgumentParser
from genericpath import exists
from logging import getLogger
from os import makedirs
from os.path import join
from re import sub
from typing import Mapping

from ..global_ import set_global_state
from ..lib.utils import clean_path, parse_bool
from ..names import UNDEF

log = getLogger(__name__)

# this can be modified during development.  it will be reset from setup.py on release.
CH2_VERSION = '0.34.0'
# new database on minor releases.  not sure this will always be a good idea.  we will see.
DB_VERSION = '-'.join(CH2_VERSION.split('.')[:2])
DB_EXTN = '.db'   # used to use .sql but auto-complete for sqlite3 didn't work

PROGNAME = 'ch2'
COMMAND = 'command'

WEB_PORT = 8000
JUPYTER_PORT = 8001

CALCULATE = 'calculate'
CONSTANTS = 'constants'
DATABASE = 'database'
DIARY = 'diary'
DUMP = 'dump'
FIT = 'fit'
FIX_FIT = 'fix-fit'
GARMIN = 'garmin'
H, HELP = 'h', 'help'
IMPORT = 'import'
JUPYTER = 'jupyter'
KIT = 'kit'
LOAD = 'load'
NO_OP = 'no-op'
PACKAGE_FIT_PROFILE = 'package-fit-profile'
READ = 'read'
SEARCH = 'search'
SHOW_SCHEDULE = 'show-schedule'
TEXT = 'text'
THUMBNAIL = 'thumbnail'
UNLOCK = 'unlock'
VALIDATE = 'validate'
WEB = 'web'

A = 'a'
ACTIVITY = 'activity'
ACTIVITY_GROUP = 'activity-group'
ACTIVITY_GROUPS = 'activity-groups'
ACTIVITY_JOURNALS = 'activity-journals'
ACTIVITY_JOURNAL_ID = 'activity-journal-id'
ACTIVITIES = 'activities'
ADD = 'add'
ADD_HEADER = 'add-header'
ADVANCED = 'advanced'
AFTER = 'after'
AFTER_BYTES = 'after-bytes'
AFTER_RECORDS = 'after-records'
ALL = 'all'
ALL_MESSAGES = 'all-messages'
ALL_FIELDS = 'all-fields'
ARG = 'arg'
BASE = 'base'
BATCH = 'batch'
BIND = 'bind'
BORDER = 'border'
CHANGE = 'change'
CHECK = 'check'
CMD = 'cmd'
COLOR = 'color'
COMPACT = 'compact'
COMPONENT = 'component'
CONSTRAINT = 'constraint'
CONTEXT = 'context'
CSV = 'csv'
D = 'd'
DARK = 'dark'
DATA = 'data'
DATE = 'date'
DEFAULT = 'default'
DEFINE = 'define'
DELETE = 'delete'
DESCRIBE = 'describe'
DESCRIPTION = 'description'
DEV = 'dev'
DIR = 'dir'
DISABLE = 'disable'
DISCARD = 'discard'
DROP = 'drop'
EMPTY = 'empty'
ENABLE = 'enable'
ENGINE = 'engine'
F = 'f'
FAST = 'fast'
FIELD = 'field'
FIELDS = 'fields'
FIX = 'fix'
FINISH = 'finish'
FIX_CHECKSUM = 'fix-checksum'
FIX_HEADER = 'fix-header'
FORMAT = 'format'
FORCE = 'force'
FTHR = 'fthr'
GREP = 'grep'
GROUP = 'group'
HEADER_SIZE = 'header-size'
HEIGHT = 'height'
INTERNAL = 'internal'
ITEM = 'item'
K = 'k'
KARG = 'karg'
LABEL = 'label'
LATITUDE = 'latitude'
LIGHT = 'light'
LIKE = 'like'
LIMIT_BYTES = 'limit-bytes'
LIMIT_RECORDS = 'limit-records'
LOG = 'log'
LOGS = 'logs'
LONGITUDE = 'longitude'
LIST = 'list'
M, MESSAGE = 'm', 'message'
MATCH = 'match'
MAX_BACK_CNT = 'max-back-cnt'
MAX_COLUMNS = 'max-columns'
MAX_COLWIDTH = 'max-colwidth'
MAX_DROP_CNT = 'max-drop-cnt'
MAX_DELTA_T = 'max-delta-t'
MAX_FWD_LEN = 'max-fwd-len'
MAX_ROWS = 'max-rows'
MAX_RECORD_LEN = 'max-record-len'
MIN_SYNC_CNT = 'min-sync-cnt'
MODEL = 'model'
MONITOR = 'monitor'
MONTH = 'month'
MONTHS = 'months'
NAME = 'name'
NAME_BAD = 'name-bad'
NAME_GOOD = 'name-good'
NAMES = 'names'
NEW = 'new'
NOT = 'not'
NOTEBOOKS = 'notebooks'
O, OUTPUT = 'o', 'output'
OFF = 'off'
OWNER = 'owner'
PASS = 'pass'
PATH = 'path'
P, PATTERN = 'p', 'pattern'
PERMANENT = 'permanent'
POSTGRESQL = 'postgresql'
PLAN = 'plan'
PORT = 'port'
PRINT = 'print'
PROFILE = 'profile'
PROFILE_VERSION = 'profile-version'
PROTOCOL_VERSION = 'protocol-version'
PROXY = 'proxy'
PWD = 'pwd'
QUERY = 'query'
RAW = 'raw'
READ_ONLY = 'read-only'
REBUILD = 'rebuild'
RECORDS = 'records'
REMOVE = 'remove'
RETIRE = 'retire'
ROOT = 'root'
RUN = 'run'
SEGMENT_JOURNALS = 'segment-journals'
SEGMENTS = 'segments'
SERVICE = 'service'
SET = 'set'
SCHEDULE = 'schedule'
SECURE = 'secure'
SHOW = 'show'
SINGLE = 'single'
SLICES = 'slices'
SOURCE = 'source'
SOURCES = 'sources'
SOURCE_ID = 'source-id'
SQLITE = 'sqlite'
START = 'start'
STATISTICS = 'statistics'
STATISTIC_NAMES = 'statistic-names'
STATISTIC_JOURNALS = 'statistic-journals'
STATUS = 'status'
STOP = 'stop'
SUB_COMMAND = 'sub-command'
SYSTEM = 'system'
TABLE = 'table'
TABLES = 'tables'
TOKENS = 'tokens'
TOPIC = 'topic'
UNDO = 'undo'
UNLIKE = 'unlike'
UNSAFE = 'unsafe'
UNSET = 'unset'
URI = 'uri'
USER = 'user'
V, VERBOSITY = 'v', 'verbosity'
VALUE = 'value'
VERSION = 'version'
W, WARN = 'w', 'warn'
WAYPOINTS = 'waypoints'
WIDTH = 'width'
WORKER = 'worker'
YEAR = 'year'
Y = 'y'


def mm(name): return '--' + name
def m(name): return '-' + name
def no(name): return 'no-%s' % name


class NamespaceWithVariables(Mapping):

    def __init__(self, ns):
        self._dict = vars(ns)

    def __getitem__(self, name):
        try:
            value = self._dict[name]
        except KeyError:
            value = self._dict[sub('-', '_', name)]
        return value

    def system_path(self, subdir=None, file=None, version=DB_VERSION, create=True):
        return base_system_path(self[BASE], subdir=subdir, file=file, version=version, create=create)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self.__dict__)


def base_system_path(base, subdir=None, file=None, version=DB_VERSION, create=True):

    from ..lib.utils import clean_path

    dir = base
    if version: dir = join(dir, version)
    if subdir: dir = join(dir, subdir)
    dir = clean_path(dir)
    if create and not exists(dir): makedirs(dir)
    if file:
        return join(dir, file)
    else:
        return dir


def color(color):
    if color.lower() not in (LIGHT, DARK, OFF):
        raise Exception(f'Bad color: {color} ({LIGHT}|{DARK}|{OFF})')
    return color


def make_parser(with_noop=False):

    from ..lib import to_date, to_time

    parser = ArgumentParser(prog=PROGNAME)

    parser.add_argument(mm(BASE), default='~/.ch2', type=clean_path, metavar='DIR',
                        help='the base directory for data (default ~/.ch2)')
    parser.add_argument(mm(READ_ONLY), action='store_true',
                        help='read-only database (so errors on write)')
    parser.add_argument(mm(LOG), metavar='FILE',
                        help='the file name for the log (command name by default)')
    parser.add_argument(mm(COLOR), type=color,
                        help=f'pretty stdout log - {LIGHT}|{DARK}|{OFF} (CAPS to save)')
    parser.add_argument(m(V), mm(VERBOSITY), default=UNDEF, type=int, metavar='N',
                        help='output level for stderr (0: silent; 5:noisy)')
    parser.add_argument(mm(DEV), action='store_true',
                        help='verbose log and stack trace on error')
    parser.add_argument(m(V.upper()), mm(VERSION), action='version', version=CH2_VERSION,
                        help='display version and exit')

    subparsers = parser.add_subparsers(title='commands', dest=COMMAND)

    # high-level commands used daily

    help = subparsers.add_parser(HELP, help='display help',
                                 description='display additional help (beyond -h) for any command')
    help.add_argument(TOPIC, nargs='?', metavar=TOPIC, help='the subject for help')

    def add_uri_options(parser, required):
        database_uri = parser.add_mutually_exclusive_group(required=required)
        database_uri.add_argument(mm(SQLITE), dest=URI, action='store_const', const=SQLITE,
                                  help='use an sqlite database with standard parameters')
        database_uri.add_argument(mm(POSTGRESQL), dest=URI, action='store_const', const=POSTGRESQL,
                                  help='use a postgresql database with standard parameters')
        database_uri.add_argument(mm(URI), help='use the given database URI')

    web = subparsers.add_parser(WEB, help='the web interface (probably all you need)',
                                description='start, stop and manage the web interface')
    web_cmds = web.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)

    def add_web_server_args(cmd, prefix='', default_address='localhost', default_port=WEB_PORT):
        if prefix: prefix += '-'
        cmd.add_argument(mm(prefix + BIND), default='localhost', metavar='ADDRESS',
                         help='bind address' + f' (default {default_address})' if default_address else '')
        cmd.add_argument(mm(prefix + PORT), default=default_port, type=int, metavar='PORT',
                         help=f'port' + f' (default {default_port})' if default_port else '')

    def add_warning_args(cmd):
        prefix = WARN + '-'
        cmd.add_argument(mm(prefix + DATA), action='store_true', help='warn user that data may be lost')
        cmd.add_argument(mm(prefix + SECURE), action='store_true', help='warn user that the system is insecure')

    web_start = web_cmds.add_parser(START, help='start the web server', description='start the web server')
    add_web_server_args(web_start, prefix=WEB)
    add_web_server_args(web_start, prefix=JUPYTER, default_port=JUPYTER_PORT)
    add_web_server_args(web_start, prefix=PROXY, default_port=None, default_address=None)
    add_warning_args(web_start)
    web_cmds.add_parser(STOP, help='stop the web server', description='stop the web server')
    web_cmds.add_parser(STATUS, help='display status of web server', description='display status of web server')
    web_service = web_cmds.add_parser(SERVICE, help='internal use only - use start/stop',
                                      description='internal use only - use start/stop')
    add_web_server_args(web_service, prefix=WEB)
    add_web_server_args(web_service, prefix=JUPYTER, default_port=JUPYTER_PORT)
    add_web_server_args(web_service, prefix=PROXY, default_port=None, default_address=None)
    add_warning_args(web_service)
    add_uri_options(web_service, False)

    read = subparsers.add_parser(READ, help='read data (also calls calculate)',
                                 description='read data (copy it to the permanent store) and '
                                             'calculate statistics (add information to the database)')
    read.add_argument(mm(FORCE), action='store_true', help='reprocess existing data')
    read.add_argument(mm(KIT), m(K), action='append', default=[], metavar='ITEM',
                      help='kit items associated with activities')
    read.add_argument(PATH, metavar='PATH', nargs='*', default=[], help='path to FIT file(s) containing data')
    read.add_argument(mm(KARG), m(K.upper()), action='append', default=[], metavar='NAME=VALUE',
                      help='keyword argument to be passed to the pipelines (can be repeated)')
    read.add_argument(mm(WORKER), metavar='ID', type=int, help='internal use only (identifies sub-process workers)')
    read.add_argument(mm(DISABLE), action='store_true', help='disable following options (they enable by default)')
    read.add_argument(mm(ACTIVITIES), action='store_true', help='enable (or disable) processing of activity data')
    read.add_argument(mm(MONITOR), action='store_true', help='enable (or disable) processing of monitor data')
    read.add_argument(mm(CALCULATE), action='store_true', help='enable (or disable) calculating statistics')

    def add_search_query(cmd, query_help='search terms (similar to SQL)'):
        cmd.add_argument(QUERY, metavar='QUERY', default=[], nargs='+', help=query_help)
        cmd.add_argument(mm(SHOW), metavar='NAME', default=[], nargs='+',
                         help='show value from matching entries')
        cmd.add_argument(mm(SET), metavar='NAME=VALUE', help='update matching entries')

    search = subparsers.add_parser(SEARCH, help='search the database', description='search the database')
    search_cmds = search.add_subparsers(title='search target', dest=SUB_COMMAND, required=True)
    search_text = search_cmds.add_parser(TEXT, help='search for text in activities',
                                         description='search for text in activities')
    add_search_query(search_text, query_help='words to search for')
    search_activities = search_cmds.add_parser(ACTIVITIES, help='search for activities',
                                               description='search for activities')
    add_search_query(search_activities)
    search_sources = search_cmds.add_parser(SOURCES, help='search for sources',
                                            description='search for sources')
    add_search_query(search_sources)

    constants = subparsers.add_parser(CONSTANTS, help='set and examine constants',
                                      description='set and examine constants')
    constants_cmds = constants.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    constants_list = constants_cmds.add_parser(LIST, help='list all constant names',
                                               description='list all constant names')
    constants_show = constants_cmds.add_parser(SHOW, help='show a value (or all values)',
                                               description='show a constant\'s value (or all values)')
    constants_show.add_argument(NAME, nargs='?', metavar='NAME', help='name (omit for all)')
    constants_show.add_argument(DATE, nargs='?', metavar='DATE',
                               help='date of value to show (omit for all)')
    constants_add = constants_cmds.add_parser(ADD, help='add a new constant name',
                                              description='add a new constant name')
    constants_add.add_argument(NAME, metavar='NAME', help='name')
    constants_add.add_argument(mm(SINGLE), action='store_true', help='allow only a single (constant) value')
    constants_add.add_argument(mm(DESCRIPTION), help='optional description')
    constants_add.add_argument(mm(VALIDATE), help='optional validation class')
    constants_set = constants_cmds.add_parser(SET, help='set or modify a value',
                                              description='set or modify a constant\'s value')
    constants_set.add_argument(NAME, metavar='NAME', help='name')
    constants_set.add_argument(VALUE, metavar='VALUE', help='value')
    constants_set.add_argument(DATE, nargs='?', metavar='DATE',
                              help='date when measured (omit for all time)')
    constants_set.add_argument(mm(FORCE), action='store_true', help='allow over-writing existing values')
    constants_unset = constants_cmds.add_parser(UNSET, help='delete a value (or all values)',
                                                description='delete a value (or all values)')
    constants_unset.add_argument(NAME, metavar='NAME', help='name')
    constants_unset.add_argument(DATE, nargs='?', metavar='DATE',
                                 help='date of value to delete (omit for all)')
    constants_unset.add_argument(mm(FORCE), action='store_true', help='allow deletion of all values')
    constants_remove = constants_cmds.add_parser(REMOVE, help='remove a constant (after deleting all values)',
                                                 description='remove a constant\'s name (after deleting all values)')
    constants_remove.add_argument(NAME, metavar='NAME', help='name')
    constants_remove.add_argument(mm(FORCE), action='store_true', help='allow remove of multiple constants')

    validate = subparsers.add_parser(VALIDATE, help='check (and optionally fix) data in the database',
                                     description='check (and optionally fix) data in the database')
    validate.add_argument(mm(FIX), action='store_true', help='correct errors when possible')

    jupyter = subparsers.add_parser(JUPYTER, help='data analysis in jupyter',
                                    description='data analysis in jupyter')
    jupyter_cmds = jupyter.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    jupyter_cmds.add_parser(LIST, help='list available templates',
                            description='list available templates and parameters for jupyter')
    jupyter_show = jupyter_cmds.add_parser(SHOW, help='display a template (starting server if needed)',
                                           description='display a template in jupyter (starting server if needed)')
    jupyter_show.add_argument(NAME, metavar='NAME', help='the template name')
    jupyter_show.add_argument(ARG, nargs='*', metavar='PARAM', help='template arguments')
    add_web_server_args(jupyter_show, prefix=JUPYTER, default_port=JUPYTER_PORT)
    add_web_server_args(jupyter_show, prefix=PROXY, default_port=None, default_address=None)
    jupyter_start = jupyter_cmds.add_parser(START, help='start a background service',
                                            description='start jupyter as a background service')
    add_web_server_args(jupyter_start, prefix=JUPYTER, default_port=JUPYTER_PORT)
    add_web_server_args(jupyter_start, prefix=PROXY, default_port=None, default_address=None)
    jupyter_cmds.add_parser(STOP, help='stop the background service',
                            description='stop the jupyter background service')
    jupyter_cmds.add_parser(STATUS, help='display status of background service')
    jupyter_service = jupyter_cmds.add_parser(SERVICE, help='internal use only - use start/stop',
                                              description='internal use only - use start/stop')
    add_web_server_args(jupyter_service, prefix=JUPYTER, default_port=JUPYTER_PORT)
    add_web_server_args(jupyter_service, prefix=PROXY, default_port=None, default_address=None)

    kit = subparsers.add_parser(KIT, help='manage kit',
                                description='add, remove, modify and display kit details')
    kit_cmds = kit.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    kit_start = kit_cmds.add_parser(START, help='define a new item (new bike, new shoe)',
                                    description='define a new item (new bike, new shoe)')
    kit_start.add_argument(GROUP, help='item group (bike, shoe, etc)')
    kit_start.add_argument(ITEM, help='item name (cotic, adidas, etc)')
    kit_start.add_argument(DATE, nargs='?', help='when created (default now)')
    kit_start.add_argument(mm(FORCE), action='store_true', help='confirm creation of a new group')
    kit_finish = kit_cmds.add_parser(FINISH, help='retire an item',
                                     description='retire an item (bike, shoe)')
    kit_finish.add_argument(ITEM, help='item name')
    kit_finish.add_argument(DATE, nargs='?', help='when to retire (default now)')
    kit_finish.add_argument(mm(FORCE), action='store_true', help='confirm change of existing date')
    kit_delete = kit_cmds.add_parser(DELETE, help='remove all entries for an item or group',
                                     description='remove all entries for an item or group')
    kit_delete.add_argument(NAME, help='item or group to delete')
    kit_delete.add_argument(mm(FORCE), action='store_true', help='confirm group deletion')
    kit_change = kit_cmds.add_parser(CHANGE, help='replace (or add) a part (wheel, innersole, etc)',
                                     description='replace (or add) a part (wheel, innersole, etc)')
    kit_change.add_argument(ITEM, help='item name (cotic, adidas, etc)')
    kit_change.add_argument(COMPONENT, help='component type (chain, laces, etc)')
    kit_change.add_argument(MODEL, help='model description')
    kit_change.add_argument(DATE, nargs='?', help='when changed (default now)')
    kit_change.add_argument(mm(FORCE), action='store_true', help='confirm creation of a new component')
    kit_change.add_argument(mm(START), action='store_true', help='set default date to start of item')
    kit_undo = kit_cmds.add_parser(UNDO, help='remove a change', description='remove a change')
    kit_undo.add_argument(ITEM, help='item name')
    kit_undo.add_argument(COMPONENT, help='component type')
    kit_undo.add_argument(MODEL, help='model description')
    kit_undo.add_argument(DATE, nargs='?', help='active date (to disambiguate models; default now)')
    kit_undo.add_argument(mm(ALL), action='store_true', help='remove all models (rather than single date)')
    kit_show = kit_cmds.add_parser(SHOW, help='display kit data',
                                   description='display kit data (show what stuff you use)')
    kit_show.add_argument(NAME, nargs='?', help='group or item to display (default all)')
    kit_show.add_argument(DATE, nargs='?', help='when to display (default now)')
    kit_show.add_argument(mm(CSV), action='store_true', help='CSV format')
    kit_statistics = kit_cmds.add_parser(STATISTICS, help='display statistics',
                                         description='display kit statistics')
    kit_statistics.add_argument(NAME, nargs='?', help='group, item, component or model')
    kit_statistics.add_argument(mm(CSV), action='store_true', help='CSV format')
    kit_rebuild = kit_cmds.add_parser(REBUILD, help='rebuild database entries')
    kit_dump = kit_cmds.add_parser(DUMP, help='dump to script')
    kit_dump.add_argument(mm(CMD), help='command to use instead of ch2')

    database = subparsers.add_parser(DATABASE, help='configure the database')
    database_cmds = database.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    database_show = database_cmds.add_parser(SHOW, help='show current configuration')
    database_list = database_cmds.add_parser(LIST, help='list available profiles')
    database_load = database_cmds.add_parser(LOAD, help='configure using the given profile')
    database_load.add_argument(mm(FORCE), action='store_true', help='overwrite existing database')
    add_uri_options(database_load, True)
    database_profiles = database_load.add_subparsers(title='profile', dest=PROFILE, required=True)
    from ..config.utils import profiles
    for name in profiles():
        database_profile = database_profiles.add_parser(name)
        database_profile.add_argument(mm(no(DIARY)), action='store_true', help='skip diary creation (for migration)')
    database_delete = database_cmds.add_parser(DELETE, help='delete the current database (or one at a given URI)')
    database_engines = database_delete.add_mutually_exclusive_group(required=False)
    database_engines.add_argument(mm(SQLITE), dest=URI, action='store_const', const=SQLITE,
                                  help='use an sqlite database with standard parameters')
    database_engines.add_argument(mm(POSTGRESQL), dest=URI, action='store_const', const=POSTGRESQL,
                                  help='use a postgresql database with standard parameters')
    database_engines.add_argument(mm(URI), help='use the given database URI')

    import_ = subparsers.add_parser(IMPORT, help='import data from a previous version')
    import_.add_argument(SOURCE, help='version or uri to import')
    import_engines = import_.add_mutually_exclusive_group(required=False)
    import_engines.add_argument(mm(SQLITE), dest=ENGINE, action='store_const', const=SQLITE,
                                help='version is an sqlite database (default as current)')
    import_engines.add_argument(mm(POSTGRESQL), dest=ENGINE, action='store_const', const=POSTGRESQL,
                                help='version is a postgresql database (default as current)')
    import_.add_argument(mm(DISABLE), action='store_true', help='disable following options (they enable by default)')
    import_.add_argument(mm(DIARY), action='store_true', help='enable (or disable) import of diary data')
    import_.add_argument(mm(ACTIVITIES), action='store_true', help='enable (or disable) import of activity data')
    import_.add_argument(mm(KIT), action='store_true', help='enable (or disable) import of kit data')
    import_.add_argument(mm(CONSTANTS), action='store_true', help='enable (or disable) import of constant data')
    import_.add_argument(mm(SEGMENTS), action='store_true', help='enable (or disable) import of segment data')

    garmin = subparsers.add_parser(GARMIN, help='download monitor data from garmin connect')
    garmin.add_argument(DIR, metavar='DIR', nargs='?', help='the directory where FIT files are stored')
    garmin.add_argument(mm(USER), metavar='USER', help='garmin connect username')
    garmin.add_argument(mm(PASS), metavar='PASSWORD', help='garmin connect password')
    garmin.add_argument(mm(DATE), metavar='DATE', type=to_date, help='date to download')
    garmin.add_argument(mm(FORCE), action='store_true', help='allow longer date range')

    calculate = subparsers.add_parser(CALCULATE, help='(re-)calculate statistics')
    calculate.add_argument(mm(FORCE), action='store_true', help='delete existing statistics')
    calculate.add_argument(mm(LIKE), action='append', default=[], metavar='PATTERN',
                           help='run only matching pipeline classes')
    calculate.add_argument(mm(UNLIKE), action='append', default=[], metavar='PATTERN',
                           help='exclude matching pipeline classes')
    calculate.add_argument(START, metavar='START', nargs='?', help='optional start date')
    calculate.add_argument(FINISH, metavar='FINISH', nargs='?', help='optional finish date (if start also given)')
    calculate.add_argument(mm(KARG), m(K.upper()), action='append', default=[], metavar='NAME=VALUE',
                           help='keyword argument to be passed to the pipelines (can be repeated)')
    calculate.add_argument(mm(WORKER), metavar='ID', type=int,
                           help='internal use only (identifies sub-process workers)')

    fit = subparsers.add_parser(FIT, help='display contents of fit file')
    fit_cmds = fit.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    fit_grep = fit_cmds.add_parser(GREP, help='show matching entries')
    fit_records = fit_cmds.add_parser(RECORDS, help='show high-level structure (ordered by time)')
    fit_tables = fit_cmds.add_parser(TABLES, help='show high-level structure (grouped in tables)')
    fit_csv = fit_cmds.add_parser(CSV, help='show high-level structure (in CSV format)')
    fit_tokens = fit_cmds.add_parser(TOKENS, help='show low-level tokens')
    fit_fields = fit_cmds.add_parser(FIELDS, help='show low-level fields (within tokens)')

    def add_fit_general(cmd):
        cmd.add_argument(mm(AFTER_RECORDS), type=int, metavar='N', default=None,
                         help='skip initial records')
        cmd.add_argument(mm(LIMIT_RECORDS), type=int, metavar='N', default=-1,
                         help='limit number of records displayed')
        cmd.add_argument(mm(AFTER_BYTES), type=int, metavar='N', default=None,
                         help='skip initial bytes')
        cmd.add_argument(mm(LIMIT_BYTES), type=int, metavar='N', default=-1,
                         help='limit number of bytes displayed')
        cmd.add_argument(m(W), mm(WARN), action='store_true', help='log additional warnings')
        cmd.add_argument(mm(no(VALIDATE)), action='store_true', help='do not validate checksum, length')
        cmd.add_argument(mm(MAX_DELTA_T), type=float, metavar='S',
                         help='validate seconds between timestamps (and non-decreasing)')
        cmd.add_argument(mm(NAME), action='store_true', help='print file name')
        cmd.add_argument(PATH, metavar='PATH', nargs='+', help='path to fit file')

    def add_fit_grep(cmd):
        cmd.add_argument(mm(NOT), action='store_true', help='print file names that don\'t match (with --name)')
        cmd.add_argument(mm(MATCH), type=int, default=-1, help='max number of matches (default -1 for all)')
        cmd.add_argument(mm(COMPACT), action='store_true', help='no space between records')
        cmd.add_argument(mm(CONTEXT), action='store_true', help='display entire record')
        cmd.add_argument(m(P), mm(PATTERN), nargs='+', metavar='MSG:FLD[=VAL]', required=True,
                         help='pattern to match (separate from PATH with --)')

    def add_fit_high_level(cmd):
        cmd.add_argument(m(M), mm(MESSAGE), nargs='+', metavar='MSG', help='display named messages')
        cmd.add_argument(m(F), mm(FIELD), nargs='+', metavar='FLD', help='display named fields')
        cmd.add_argument(mm(INTERNAL), action='store_true', help='display internal messages')

    def add_fit_very_high_level(cmd):
        cmd.add_argument(mm(ALL_MESSAGES), action='store_true', help='display undocumented messages')
        cmd.add_argument(mm(ALL_FIELDS), action='store_true', help='display undocumented fields')

    for cmd in fit_grep, fit_records, fit_tables, fit_csv, fit_tokens, fit_fields:
        add_fit_general(cmd)

    for cmd in fit_records, fit_tables, fit_csv:
        add_fit_high_level(cmd)

    for cmd in fit_records, fit_tables:
        add_fit_very_high_level(cmd)

    add_fit_grep(fit_grep)

    for cmd in fit_grep, fit_records, fit_tables:
        cmd.add_argument(mm(WIDTH), type=int,
                         help='display width')

    fix_fit = subparsers.add_parser(FIX_FIT, help='fix a corrupted fit file')
    fix_fit.add_argument(PATH, metavar='PATH', nargs='+', help='path to fit file')
    fix_fit.add_argument(m(W), mm(WARN), action='store_true', help='additional warning messages')
    fix_fit_output = fix_fit.add_argument_group(title='output (default hex to stdout)').add_mutually_exclusive_group()
    fix_fit_output.add_argument(m(O), mm(OUTPUT), help='output file for fixed data (otherwise, stdout)')
    fix_fit_output.add_argument(mm(DISCARD), action='store_true', help='discard output (otherwise, stdout)')
    fix_fit_output.add_argument(mm(RAW), action='store_true', help='raw binary to stdout (otherwise, hex encoded)')
    fix_fit_output.add_argument(mm(NAME_BAD), action='store_false', dest=NAME, default=None,
                                help='print file name if bad')
    fix_fit_output.add_argument(mm(NAME_GOOD), action='store_true', dest=NAME, default=None,
                                help='print file name if good')
    fix_fit_process = fix_fit.add_argument_group(title='processing (default disabled)')
    fix_fit_process.add_argument(mm(ADD_HEADER), action='store_true', help='preprend a new header')
    fix_fit_stage = fix_fit_process.add_mutually_exclusive_group()
    fix_fit_stage.add_argument(mm(DROP), action='store_true',
                               help='search for data that can be dropped to give a successful parse')
    fix_fit_stage.add_argument(mm(SLICES), metavar='A:B,C:D,...', help='data slices to pick')
    fix_fit_stage.add_argument(mm(START), type=to_time, metavar='TIME', help='change start time')
    fix_fit_process.add_argument(mm(FIX_HEADER), action='store_true', help='modify the header')
    fix_fit_process.add_argument(mm(FIX_CHECKSUM), action='store_true', help='modify the checksum')
    fix_fit_process.add_argument(mm(no(FORCE)), action='store_false', dest=FORCE,
                                 help='don\'t parse record contents')
    fix_fit_process.add_argument(mm(no(VALIDATE)), action='store_false', dest=VALIDATE,
                                 help='don\'t validate the final data')
    fix_fit_params = fix_fit.add_argument_group(title='parameters')
    fix_fit_params.add_argument(mm(HEADER_SIZE), type=int, metavar='N',
                                help='header size (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROTOCOL_VERSION), type=int, metavar='N',
                                help='protocol version (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROFILE_VERSION), type=int, metavar='N',
                                help='profile version (validation and/or new header)')
    fix_fit_params.add_argument(mm(MIN_SYNC_CNT), type=int, metavar='N', default=3,
                                help='minimum number of records to read for synchronization')
    fix_fit_params.add_argument(mm(MAX_RECORD_LEN), type=int, metavar='N', default=None,
                                help='maximum record length')
    fix_fit_params.add_argument(mm(MAX_DROP_CNT), type=int, metavar='N', default=1,
                                help='maximum number of gaps to drop')
    fix_fit_params.add_argument(mm(MAX_BACK_CNT), type=int, metavar='N', default=3,
                                help='maximum number of readable records to discard in a single gap')
    fix_fit_params.add_argument(mm(MAX_FWD_LEN), type=int, metavar='N', default=200,
                                help='maximum number of bytes to drop in a single gap')
    fix_fit_params.add_argument(mm(MAX_DELTA_T), type=float, metavar='S',
                                help='max number of seconds between timestamps')

    thumbnail = subparsers.add_parser(THUMBNAIL,
                                      help='generate a thumbnail map of an activity')
    thumbnail.add_argument(ACTIVITY, metavar='ACTIVITY',
                           help='an activity ID or date')

    if with_noop:
        noop = subparsers.add_parser(NO_OP,
                                     help='used within jupyter (no-op from cmd line)')

    package_fit_profile = subparsers.add_parser(PACKAGE_FIT_PROFILE,
                                                help='parse and save the global fit profile (dev only)')
    package_fit_profile.add_argument(PATH, metavar='PROFILE',
                                     help='the path to the profile (Profile.xlsx)')
    package_fit_profile.add_argument(m(W), mm(WARN), action='store_true',
                                     help='additional warning messages')

    show_schedule = subparsers.add_parser(SHOW_SCHEDULE, help='print schedule locations in a calendar')
    show_schedule.add_argument(SCHEDULE, metavar='SCHEDULE', help='schedule to test')
    show_schedule.add_argument(mm(START), metavar='DATE', help='date to start displaying data')
    show_schedule.add_argument(mm(MONTHS), metavar='N', type=int, help='number of months to display')

    unlock = subparsers.add_parser(UNLOCK, help='remove database locking')

    return parser


def bootstrap_dir(base, *args, configurator=None, post_config=None):
    # used in tests, given a base directory

    from ..sql.database import sqlite_uri
    from ..sql.system import SystemConstant

    args = [mm(BASE), base] + list(args)
    parser = make_parser()
    ns = NamespaceWithVariables(parser.parse_args(args=args))
    if configurator:
        data = set_global_state(ns)
        data.sys.set_constant(SystemConstant.DB_URI, sqlite_uri(data.base), force=True)
        with data.db.session_context() as s:
            configurator(s, data.base)
    args += post_config if post_config else []
    ns = NamespaceWithVariables(parser.parse_args(args=args))
    data = set_global_state(ns)
    data.sys.set_constant(SystemConstant.DB_URI, sqlite_uri(data.base), force=True)
    return ns, data


def parse_pairs(pairs, convert=True, multi=False, comma=False):

    from ..lib import to_time

    # simple name, value pairs. owner and constraint supplied by command.
    d = {}
    if pairs is not None:
        for pair in pairs:
            name, value = pair.split('=', 1)
            if convert:
                for type in (int, float, to_time, lambda x: parse_bool(x, default=None)):
                    try:
                        value = type(value)
                        break
                    except ValueError:
                        pass
            if multi:
                if name not in d:
                    d[name] = []
                d[name].append(value)
            elif comma:
                if name in d:
                    d[name] = d[name] + ',' + value
                else:
                    d[name] = value
            else:
                d[name] = value
    return d


def infer_flags(args, *names):
    flags = {name: args[name] for name in names}
    # if none were given, all are assumed
    if all(not flags[name] for name in names):
        for name in names: flags[name] = True
    if args[DISABLE]:
        for name in flags: flags[name] = not flags[name]
    return flags






from abc import abstractmethod
from contextlib import nullcontext
from logging import getLogger

from psutil import cpu_count
from sqlalchemy import text
from sqlalchemy.sql.functions import count

from .loader import SqliteLoader, PostgresqlLoader
from ..commands.args import SQLITE, POSTGRESQL, BATCH, mm, KARG
from ..lib.utils import timing
from ..lib.workers import ProgressTree, Workers
from ..sql import Pipeline, SystemConstant, Interval, PipelineType, StatisticJournal
from ..sql.database import scheme
from ..sql.types import short_cls

log = getLogger(__name__)

CPU_FRACTION = 0.9
MAX_REPEAT = 3


def count_statistics(s):
    return s.query(count(StatisticJournal.id)).scalar()


def run_pipeline(data, type, like=tuple(), unlike=tuple(), id=None, progress=None, **extra_kargs):
    with data.db.session_context() as s:
        if id is None:  # don't run for each worker
            if type in (PipelineType.CALCULATE, PipelineType.READ_ACTIVITY, PipelineType.READ_MONITOR):
                Interval.clean(s)
        local_progress = ProgressTree(Pipeline.count(s, type, like=like, unlike=unlike, id=id), parent=progress)
        for pipeline in Pipeline.all(s, type, like=like, unlike=unlike, id=id):
            kargs = dict(pipeline.kargs)
            kargs.update(extra_kargs)
            msg = f'Ran {short_cls(pipeline.cls)}'
            if 'activity_group' in kargs: msg += f' ({kargs["activity_group"]})'
            log.debug(f'Running {pipeline.cls}({pipeline.args}, {kargs})')
            with timing(msg):
                before = None if id else count_statistics(s)
                pipeline.cls(data, *pipeline.args, id=pipeline.id, progress=local_progress, **kargs).run()
                after = None if id else count_statistics(s)
            if before or after:
                log.info(f'{msg}: statistic count {before} -> {after} (change of {after - before})')


class BasePipeline:

    def __init__(self, *args, **kargs):
        if args or kargs:
            log.warning(f'Unused pipeline argument(s) for {short_cls(self)}: {args} {list(kargs.keys())}')

    def _assert(self, name, value):
        if value is None:
            raise Exception(f'Undefined {name}')
        else:
            return value


class MultiProcPipeline(BasePipeline):

    def __init__(self, data, *args, owner_out=None, force=False, progress=None,
                 overhead=1, cost_calc=20, cost_write=1, n_cpu=None, worker=None, id=None, **kargs):
        self._data = data
        self.owner_out = owner_out or self  # the future owner of any calculated statistics
        self.force = force  # force re-processing
        self.__progress = progress
        self.overhead = overhead  # next three args are used to decide if workers are needed
        self.cost_calc = cost_calc  # see _cost_benefit for full details
        self.cost_write = cost_write  # defaults guarantee a single thread
        self.n_cpu = max(1, int(cpu_count() * CPU_FRACTION)) if n_cpu is None else n_cpu  # number of cpus available
        self.worker = worker  # if True, then we're in a sub-process
        self.id = id  # the id for the pipeline entry in the database (passed to sub-processes)
        super().__init__(*args, **kargs)

    def run(self):
        with self._data.db.session_context() as s:
            self._startup(s)

            if self.force:
                if self.worker:
                    log.warning('Worker deleting data')
                self._delete(s)

            missing = self._missing(s)
            log.debug(f'Have {len(missing)} missing ranges')

            if self.worker:
                log.debug('Worker, so execute directly')
                self._run_all(s, missing, None)
            else:
                local_progress = ProgressTree(len(missing), parent=self.__progress)
                if not missing:
                    log.info(f'No missing data for {short_cls(self)}')
                    local_progress.complete()
                else:
                    self.__flush_wal(s)
                    n_total, n_parallel = self.__cost_benefit(missing, self.n_cpu)
                    if n_parallel < 2 or len(missing) == 1:
                        self._run_all(s, missing, local_progress)
                    else:
                        self.__spawn(s, missing, n_total, n_parallel, local_progress)
            self._shutdown(s)

    def __flush_wal(self, s):
        s.commit()
        if str(s.get_bind().url).startswith(SQLITE):
            log.debug('Clearing WAL')
            s.execute(text('pragma wal_checkpoint(RESTART);'))
            s.commit()
            log.debug('Cleared WAL')

    def _run_all(self, s, missing, progress=None):
        local_progress = progress.increment_or_complete if progress else nullcontext
        for missed in missing:
            with local_progress():
                self._run_one(s, missed)
                s.commit()

    def _startup(self, s):
        pass

    def _shutdown(self, s):
        pass

    # as a general rule, _missing and _args should be implemented together
    @abstractmethod
    def _missing(self, s):
        raise NotImplementedError()

    @abstractmethod
    def _delete(self, s):
        raise NotImplementedError()

    @abstractmethod
    def _run_one(self, s, missed):
        raise NotImplementedError()

    def __cost_benefit(self, missing, n_cpu):

        # is it worth using workers?  there's some cost in starting them up and there will be contention
        # in accessing the database.
        # let's say COST (for one missing time) is COST_WRITE + COST_CALC (ignoring units).
        # if we have N_MISSING tasks divided into N_TOTAL workloads amongst N_PARALLEL workers
        # then we have these conditions (in order):
        #   N_PARALLEL * (COST_WRITES/ COST) <= 1 so that we avoid blocking completely on writes
        #   N_PARALLEL <= N_CPU
        #   N_TOTAL <= N_MISSING / N_PARALLEL
        #   (N_MISSING / N_TOTAL) * COST > OVERHEAD so we're not wasting our time
        #   N_TOTAL <= N_PARALLEL * MAX_REPEAT because we want large batches, but not too large
        # really we should include estimates of disk and cpu speed here, in which case we need to separate
        # out COST_READ too (currently folded into COST_CALC).

        log.debug(f'Batching for n_cpu={n_cpu}, overhead={self.overhead}, '
                  f'cost_writes={self.cost_write}, cost_calc={self.cost_calc}')
        n_missing = len(missing)
        cost = self.cost_write + self.cost_calc
        limit = cost / self.cost_write
        log.debug(f'Limit on parallel workers from database contention is {limit:3.1f}')
        log.debug(f'Limit on parallel workers from CPU count is {n_cpu:d}')
        n_parallel = int(min(limit, n_cpu))
        n_total = int((n_missing + n_parallel - 1) / n_parallel)
        log.debug(f'Limit on total workers from work available is {n_total:d}')
        limit = cost * n_missing / self.overhead
        log.debug(f'Limit on total workers from overhead is {limit:3.1f}')
        n_total = min(n_total, int(limit))
        limit = n_parallel * MAX_REPEAT
        log.debug(f'Limit on total workers to boost batch size is {limit:d}')
        n_total = min(n_total, limit)
        log.info(f'Threads: {n_total}/{n_parallel}')
        return n_total, n_parallel

    def __spawn(self, s, missing, n_total, n_parallel, progress):

        # unfortunately we have to do things with contiguous dates, which may introduce systematic
        # errors in our timing estimates

        n_missing = len(missing)
        workers = Workers(self._data, n_parallel, self.owner_out, self._base_command())
        start, finish = None, -1
        for i in range(n_total):
            start = finish + 1
            finish = int(0.5 + (i+1) * (n_missing-1) / n_total)
            if start > finish: raise Exception('Bad chunking logic')
            with progress.increment_or_complete(finish - start + 1):
                workers.run(self.id, self._args(missing, start, finish))

        workers.wait()

    # as a general rule, _missing and _args should be implemented together
    @abstractmethod
    def _args(self, missing, start, finish):
        raise NotImplementedError()

    @abstractmethod
    def _base_command(self):
        raise NotImplementedError()


class UniProcPipeline(MultiProcPipeline):

    def __init__(self, *args, overhead=None, cost_calc=None, cost_write=None, n_cpu=None, worker=None, id=None,
                 **kargs):
        super().__init__(*args, overhead=1, cost_calc=0, cost_write=1, n_cpu=None, worker=None, id=None, **kargs)


class LoaderMixin:

    loaders = {SQLITE: SqliteLoader, POSTGRESQL: PostgresqlLoader}

    def __init__(self, *args, batch=True, **kargs):
        super().__init__(*args, **kargs)
        self.__batch = batch

    def _base_command(self):
        cmd = super()._base_command()
        if not self.__batch: cmd += f' {mm(KARG)} {BATCH}={False}'
        return cmd

    def _get_loader(self, s, add_serial=None, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        if add_serial is None:
            raise Exception('Select serial use')
        else:
            kargs['add_serial'] = add_serial
        scheme_ = scheme(self._data.sys.get_constant(SystemConstant.DB_URI))
        if scheme_ == POSTGRESQL and 'batch' not in kargs:
            kargs['batch'] = self.__batch
            self.__batch = False  # only set once or we get multiple callbacks
        if scheme_ in self.loaders:
            log.debug(f'Using loader for {scheme_}')
            return self.loaders[scheme_](s, **kargs)
        else:
            raise Exception(f'Unknown scheme: {scheme_}')


class OwnerInMixin:

    def __init__(self, *args, owner_in=None, **kargs):
        self.owner_in = self._assert('owner_in', owner_in)
        super().__init__(*args, **kargs)




import datetime as dt
from enum import IntEnum
from logging import getLogger

from sqlalchemy import Column, Integer, ForeignKey, Text, UniqueConstraint, Float, desc, asc, Index
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import relationship, backref, synonym
from sqlalchemy.orm.exc import NoResultFound

from .source import Interval
from ..support import Base
from ..types import Time, ShortCls, Name, name_and_title
from ..utils import add
from ...diary.model import TYPE, MEASURES, SCHEDULES
from ...lib.date import format_seconds, local_date_to_time, time_to_local_time
from ...lib.utils import sigfig
from ...names import Units, simple_name

log = getLogger(__name__)


class StatisticName(Base):

    __tablename__ = 'statistic_name'

    id = Column(Integer, primary_key=True)
    # having name and title is worryingly redundant.  the justification is that name is the thing exposed
    # at the command line, so should be easy to type and parse.  the title is for the GUI.  unfortunately,
    # currently, names are used in search in the GUI.  also, arguably, this should not affect the schema,
    # however the mapping from title to name is not reversible.
    name = Column(Name, nullable=False)  # no need for index - can use unique(name, owner)
    title = Column(Text, nullable=False)
    description = Column(Text)
    units = Column(Text)
    summary = Column(Text)  # max, min, etc - comma separated list
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    statistic_journal_type = Column(Integer, nullable=False)  # StatisticJournalType
    UniqueConstraint(name, owner)

    def __init__(self, **kargs):
        super().__init__(**name_and_title(kargs))

    def __str__(self):
        return '%s.%s' % (self.owner, self.name)

    @property
    def summaries(self):
        if self.summary.strip():
            return [simple_name(x) for x in self.summary.split(',')]
        else:
            return []

    def qualified_name(self, s, activity_group):
        from . import ActivityGroup
        activity_group = ActivityGroup.from_name(s, activity_group)
        if activity_group:
            return f'{self.name}:{activity_group.name}'
        else:
            return self.name

    @classmethod
    def add_if_missing(cls, s, name, type, units, summary, owner, description=None, title=None):
        s.commit()  # start new transaction here in case rollback
        q = s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner)
        statistic_name = q.one_or_none()
        if not statistic_name:
            statistic_name = add(s, StatisticName(name=name, units=units, summary=summary, owner=owner,
                                                  statistic_journal_type=type, description=description,
                                                  title=title))
            try:
                s.flush()
            except IntegrityError as e:  # worker may have created in parallel, so read
                log.debug(f'Rollback for {e}')
                s.rollback()
                log.debug('Now trying retrieval...')
                statistic_name = q.one()
                log.debug('Retrieved')
        else:
            if statistic_name.statistic_journal_type != type:
                raise Exception('Changing type on %s (%s -> %s)' %
                                (statistic_name.name, statistic_name.statistic_journal_type, type))
            if statistic_name.units != units:
                log.warning('Changing units on %s (%s -> %s)' % (statistic_name.name, statistic_name.units, units))
                statistic_name.units = units
                s.flush()
            if statistic_name.summary != summary:
                log.warning('Changing summary on %s (%s -> %s)' % (statistic_name.name, statistic_name.summary, summary))
                statistic_name.summary = summary
                s.flush()
            if statistic_name.description != description:
                log.warning('Changing description on %s (%s -> %s)' % (statistic_name.name, statistic_name.description, description))
                statistic_name.description = description
                s.flush()
        return statistic_name

    @classmethod
    def from_name(cls, s, name, owner):
        try:
            return s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == owner).one()
        except NoResultFound:
            raise Exception(f'Statistic name {name} for owner {owner} was not found')

    @classmethod
    def parse(cls, qname, default_owner=None, default_activity_group=None):
        '''
        This parses the standard, extended format for naming statistics.
        It has the form Ownwer.name:group where:
        Owner is the short_cls that created the value and provides a namespace (optional);
        name is the simple_name that identifies the value (required);
        group is the activity group name (optional).
        A null/None group is specified by Owner.name: (ie a trailing colon).
        If an owner is not specified then 'any' owner is used (risk of ambiguity).
        If a group is not specified then the meaning depends on the context (any or default).
        '''
        if ':' in qname:
            left, group = qname.rsplit(':', 1)
        else:
            left, group = qname, default_activity_group
        if '.' in left:
            owner, name = left.rsplit('.', 1)
        else:
            owner, name = default_owner, left
        log.debug(f'Parsed {qname} as {owner}.{name}:{group}')
        return owner, name, group


class StatisticJournalType(IntEnum):

    STATISTIC = 0
    INTEGER = 1
    FLOAT = 2
    TEXT = 3
    TIMESTAMP = 4


class StatisticJournal(Base):

    __tablename__ = 'statistic_journal'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast delete of subtypes
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'),
                               nullable=False, index=True)
    statistic_name = relationship('StatisticName')
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                       nullable=False, index=True)
    source = relationship('Source')
    time = Column(Time, nullable=False)
    # serial "counts" along values in the timeseries.  it's optional.  for garmin, all values appear each
    # record, so all imported values share the same serial.  but that's not true for the corrected elevation,
    # for example.
    serial = Column(Integer)

    UniqueConstraint(statistic_name_id, time, source_id)
    UniqueConstraint(serial, source_id, statistic_name_id)
    Index('from_activity_timespan', source_id, statistic_name_id, time)  # time last since inequality

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.STATISTIC,
        'polymorphic_on': 'type'
    }

    def __str__(self):
        try:
            return 'StatisticJournal "%s"' % self.value
        except AttributeError:
            return 'StatisticJournal base'

    @classmethod
    def before_flush(cls, s):
        cls.__delete_null_values(s)

    @classmethod
    def __delete_null_values(cls, s):
        # drop null values.  these are instances that were created when reading missing values from the diary
        # for display.  the actual values written are done in a separate thread (and have non-null values).
        # (it's difficult to do this by discarding after reading because of auto-flush)
        for instance in s.new:
            if isinstance(instance, StatisticJournal) and hasattr(instance, 'value') and instance.value is None:
                s.expunge(instance)

    @classmethod
    def add(cls, s, name, units, summary, owner, source, value, time, serial, type,
            description=None):
        statistic_name = StatisticName.add_if_missing(s, name, type, units, summary, owner,
                                                      description=description)
        journal = STATISTIC_JOURNAL_CLASSES[type](statistic_name=statistic_name, source=source,
                                                  value=value, time=time, serial=serial)
        s.add(journal)
        return journal

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%d' % self.value
        elif units == Units.M:
            if self.value > 2000:
                return '%d km' % (self.value / 1000)
            else:
                return '%d m' % self.value
        elif units == Units.S:
            return format_seconds(self.value)
        elif units in (Units.KMH, Units.PC, Units.BPM, Units.STEPS_UNITS, Units.W, Units.KJ):
            return '%d %s' % (self.value, units)
        elif units == Units.KCAL:
            return '%s %s' % (sigfig(self.value, 2), units)
        else:
            return '%d %s' % (self.value, units)

    def measures_as_text(self, date):
        words = []
        if hasattr(self, 'measures'):
            for measure in sorted(self.measures,
                                  key=lambda measure: measure.source.schedule.frame_length_in_days(date)):
                if words:
                    words += [',']
                quintile = 1 + min(4, measure.percentile / 20)
                words += [('quintile-%d' % quintile, '%d%%' % int(measure.percentile))]
                if measure.rank < 5:
                    words += [':', ('rank-%d' % measure.rank, '%d' % measure.rank)]
                words += ['/' + measure.source.schedule.describe(compact=True)]
        return words

    def measures_as_model(self, date):
        if hasattr(self, 'measures'):
            measures = {TYPE: MEASURES, SCHEDULES: {}}
            for measure in sorted(self.measures,
                                  key=lambda measure: measure.source.schedule.frame_length_in_days(date),
                                  reverse=True):
                measures[SCHEDULES][measure.source.schedule.describe(compact=True)] = (measure.percentile, measure.rank)
            return measures
        else:
            return None

    @classmethod
    def at_date(cls, s, date, name, owner, activity_group, source_id=None):
        from . import ActivityGroup, Source
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        q = s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   StatisticJournal.time >= start,
                   StatisticJournal.time < finish,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group))
        if source_id is not None:
            q = q.filter(StatisticJournal.source_id == source_id)
        return q.one_or_none()

    @classmethod
    def at(cls, s, time, name, owner, activity_group):
        from . import ActivityGroup, Source
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   StatisticJournal.time == time,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group)).one_or_none()

    @classmethod
    def at_like(cls, s, time, name, owner, activity_group):
        from . import ActivityGroup, Source
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name.like(name),
                   StatisticJournal.time == time,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group)).all()

    @classmethod
    def before(cls, s, time, name, owner, activity_group):
        from . import ActivityGroup, Source
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   StatisticJournal.time <= time,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group)). \
            order_by(desc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def before_not_null(cls, s, time, name, owner, activity_group):
        # this is complex because we need the concrete statistic type to get the value
        from . import ActivityGroup, Source
        try:
            statistic_name = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == owner).one()
        except NoResultFound:
            raise Exception(f'The statistic name {owner}.{name} is undefined in the database')
        journal = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]
        return s.query(journal).join(Source). \
            filter(journal.statistic_name == statistic_name,
                   journal.time <= time,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group),
                   journal.value != None). \
            order_by(desc(journal.time)).limit(1).one_or_none()

    @classmethod
    def after(cls, s, time, name, owner, activity_group):
        from . import ActivityGroup, Source
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   StatisticJournal.time >= time,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group)). \
            order_by(asc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def at_interval(cls, s, start, schedule, statistic_owner, statistic_name, interval_owner):
        return s.query(StatisticJournal).join(StatisticName, Interval). \
                    filter(StatisticJournal.statistic_name_id == StatisticName.id,
                           Interval.schedule == schedule,
                           Interval.start == start,
                           Interval.owner == interval_owner,
                           StatisticName.owner == statistic_owner,
                           StatisticName.id == statistic_name.id). \
                    order_by(Interval.activity_group_id,  # order places summary stats from same source together
                             StatisticName.name).all()

    @classmethod
    def for_source(cls, s, source_id, name, owner, activity_group):
        from . import ActivityGroup, Source
        return s.query(StatisticJournal).join(StatisticName, Source). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   Source.activity_group == ActivityGroup.from_name(s, activity_group),
                   StatisticJournal.source_id == source_id).one_or_none()


class StatisticJournalInteger(StatisticJournal):

    __tablename__ = 'statistic_journal_integer'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Integer, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.INTEGER
    }

    def set(self, value):
        self.value = value if value is None else int(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, source, value, time, serial=None, description=None):
        return super().add(s, name, units, summary, owner, source, value, time, serial,
                           StatisticJournalType.INTEGER, description=description)


class StatisticJournalFloat(StatisticJournal):

    __tablename__ = 'statistic_journal_float'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Float, nullable=False)

    def set(self, value):
        self.value = value if value is None else float(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, source, value, time, serial=None, description=None):
        return super().add(s, name, units, summary, owner, source, value, time, serial,
                           StatisticJournalType.FLOAT, description=description)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.FLOAT
    }

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%f' % self.value
        elif units == Units.M:
            if self.value > 2000:
                return '%.1f km' % (self.value / 1000)
            else:
                return '%d m' % int(self.value)
        elif units == Units.KM:
            if self.value > 2:
                return '%.1f km' % self.value
            else:
                return '%d m' % int(self.value * 1000)
        elif units == Units.S:
            return format_seconds(self.value)
        elif units in (Units.KMH, Units.PC, Units.KG, Units.W, Units.KJ):
            return '%.1f %s' % (self.value, units)
        elif units == Units.KCAL:
            return '%s %s' % (sigfig(self.value, 2), units)
        elif units in (Units.BPM, Units.STEPS_UNITS):
            return '%d %s' % (int(self.value), units)
        elif units == Units.FF:
            return '%d' % int(self.value)
        else:
            return '%s %s' % (self.value, units)


class StatisticJournalText(StatisticJournal):

    __tablename__ = 'statistic_journal_text'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Text, nullable=False)

    def set(self, value):
        self.value = value if value is None else str(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, source, value, time, serial=None, description=None):
        return super().add(s, name, units, summary, owner, source, value, time, serial,
                           StatisticJournalType.TEXT, description=description)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.TEXT
    }

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%s' % self.value
        else:
            return '%s %s' % (self.value, units)


class StatisticJournalTimestamp(StatisticJournal):

    __tablename__ = 'statistic_journal_timestamp'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = synonym('time')

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.TIMESTAMP
    }

    @classmethod
    def add(cls, s, name, units, summary, owner, source, time, serial=None, description=None):
        statistic_name = StatisticName.add_if_missing(s, name, StatisticJournalType.TIMESTAMP,
                                                      units, summary, owner, description=description)
        journal = StatisticJournalTimestamp(statistic_name=statistic_name, source=source, time=time, serial=serial)
        s.add(journal)
        return journal

    def formatted(self):
        return time_to_local_time(self.value)


class StatisticMeasure(Base):

    __tablename__ = 'statistic_measure'

    id = Column(Integer, primary_key=True)
    statistic_journal_id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'),
                                  nullable=False, index=True)
    statistic_journal = relationship('StatisticJournal',
                                     backref=backref('measures', cascade='all, delete-orphan',
                                                     passive_deletes=True,
                                                     order_by='desc(StatisticMeasure.rank)'))
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                       nullable=False, index=True)  # must be an interval
    source = relationship('Source')
    rank = Column(Integer, nullable=False)  # 1 is best [1..n]
    percentile = Column(Float, nullable=False)  # 100 is best [0..100]
    quartile = Column(Integer)  # 0..4 at the min, 25%, median, 75% and max points


STATISTIC_JOURNAL_CLASSES = {
    StatisticJournalType.INTEGER: StatisticJournalInteger,
    StatisticJournalType.FLOAT: StatisticJournalFloat,
    StatisticJournalType.TEXT: StatisticJournalText,
    StatisticJournalType.TIMESTAMP: StatisticJournalTimestamp
}

STATISTIC_JOURNAL_TYPES = {
    StatisticJournalInteger: StatisticJournalType.INTEGER,
    StatisticJournalFloat: StatisticJournalType.FLOAT,
    StatisticJournalText: StatisticJournalType.TEXT,
    StatisticJournalTimestamp: StatisticJournalType.TIMESTAMP
}

TYPE_TO_JOURNAL_CLASS = {
    int: StatisticJournalInteger,
    float: StatisticJournalFloat,
    str: StatisticJournalText,
    dt.datetime: StatisticJournalTimestamp
}



from logging import getLogger, NullHandler

from sys import version_info

getLogger('bokeh').addHandler(NullHandler())
getLogger('tornado').addHandler(NullHandler())


class FatalException(Exception):

    '''
    Base class for exceptions that we can't ignore at some higher level
    (fundamental things like bad config).
    '''

    pass


from .commands.args import COMMAND, make_parser, NamespaceWithVariables, PROGNAME, HELP, DEV, DIARY, FIT, \
    PACKAGE_FIT_PROFILE, ACTIVITIES, NO_OP, DATABASE, CONSTANTS, CALCULATE, SHOW_SCHEDULE, MONITOR, GARMIN, \
    UNLOCK, DUMP, FIX_FIT, CH2_VERSION, JUPYTER, KIT, WEB, READ, IMPORT, THUMBNAIL, CHECK, SEARCH, VALIDATE, BASE
from .global_ import set_global_dev, set_global_data, set_global_state
from .commands.constants import constants
from .commands.validate import validate
from .commands.database import database
from .commands.fit import fit
from .commands.fix_fit import fix_fit
from .commands.garmin import garmin
from .commands.help import help, Markdown
from .commands.import_ import import_
from .commands.jupyter import jupyter
from .commands.kit import kit
from .commands.package_fit_profile import package_fit_profile
from .commands.search import search
from .commands.calculate import calculate
from .commands.show_schedule import show_schedule
from .commands.thumbnail import thumbnail
from .commands.unlock import unlock
from .commands.read import read
from .commands.web import web
from .lib.log import make_log_from_args, set_log_color
from .sql.database import SystemConstant
from .sql.system import System, Data

log = getLogger(__name__)


def no_op(args, system, db):
    '''
## no-op

This is used internally when accessing data in Jupyter or configuring the system
at the command line.
    '''
    pass


COMMANDS = {CONSTANTS: constants,
            DATABASE: database,
            FIT: fit,
            FIX_FIT: fix_fit,
            GARMIN: garmin,
            HELP: help,
            IMPORT: import_,
            JUPYTER: jupyter,
            KIT: kit,
            CALCULATE: calculate,
            NO_OP: no_op,
            PACKAGE_FIT_PROFILE: package_fit_profile,
            READ: read,
            SEARCH: search,
            SHOW_SCHEDULE: show_schedule,
            THUMBNAIL: thumbnail,
            UNLOCK: unlock,
            VALIDATE: validate,
            WEB: web
            }


def args_and_command():
    parser = make_parser()
    ns = parser.parse_args()
    command_name = ns.command if hasattr(ns, COMMAND) else None
    command = COMMANDS[command_name] if command_name in COMMANDS else None
    if command_name == NO_OP: ns.verbose = 0
    args = NamespaceWithVariables(ns)
    return args, command, command_name


def versions():
    log.info('Version %s' % CH2_VERSION)
    if version_info < (3, 7):
        raise Exception('Please user Python 3.7 or more recent')


def main():
    versions()
    args, command, command_name = args_and_command()
    data = set_global_state(args)
    try:
        if not command:
            log.debug('If you are seeing the "No command given" error during development ' +
                      'you may have forgotten to set the command name via `set_defaults()`.')
            raise Exception('No command given (try `ch2 help`)')
        elif command_name not in (DATABASE, PACKAGE_FIT_PROFILE, HELP):
            db = data.db if command_name not in (PACKAGE_FIT_PROFILE, HELP) else None
            if not db:
                refuse_until_configured(command_name, False)
            elif db.no_data():
                refuse_until_configured(command_name, True)
        command(args, data)
    except KeyboardInterrupt:
        log.critical('User abort')
        exit(1)
    except Exception as e:
        log.critical(e)
        log.info('See `%s %s` for available commands.' % (PROGNAME, HELP))
        log.info('Docs at http://andrewcooke.github.io/choochoo')
        if not args or args[DEV]:
            raise
        exit(2)


def refuse_until_configured(command_name, uri):
    Markdown().print(f'''
Welcome to Choochoo.

You must configure the database before use (no {"schema" if uri else "uri"}).

Please use the {PROGNAME} {DATABASE} command.
''')
    if command_name != WEB:
        exit(3)




from logging import getLogger
from os import unlink

from sqlalchemy_utils import create_database, drop_database, database_exists
from uritools import urisplit

from .args import mm, SUB_COMMAND, LIST, PROFILE, BASE, SHOW, DB_VERSION, URI, SQLITE, POSTGRESQL, \
    FORCE, DELETE
from .help import Markdown
from ..config.utils import profiles, get_profile
from ..lib import log_current_exception
from ..lib.utils import clean_path
from ..sql import SystemConstant
from ..sql.database import sqlite_uri, postgresql_uri

log = getLogger(__name__)


def database(args, data):
    '''
## database

    > ch2 database load (--sqlite|--pgsql|--uri URI) [--delete] PROFILE

Load the initial database schema.

    > ch2 database list

List the available profiles.

    > ch2 database show

Show the current database state.

    > ch2 database delete

Delete the current database.
    '''
    action = args[SUB_COMMAND]
    if action == SHOW:
        show(data.sys)
    elif action == LIST:
        list()
    elif action == DELETE:
        uri = args[URI] or sys.get_constant(SystemConstant.DB_URI, none=True)
        if not uri: raise Exception('No current database is defined')
        delete(uri, data.sys)
    else:
        load(data.sys, args[BASE], args[PROFILE], args[URI], args[FORCE])


def database_really_exists(uri):
    try:
        return database_exists(uri)
    except Exception:
        log_current_exception(traceback=False)
        return False


def show(sys):
    uri = sys.get_constant(SystemConstant.DB_URI, none=True)
    version = sys.get_constant(SystemConstant.DB_VERSION, none=True)
    if uri:
        print(f'{URI}:     {uri}')
        print(f'version: {version}')
        print(f'exists:  {database_really_exists(uri)}')
    else:
        print('no database configured')
    return


def list():
    fmt = Markdown()
    for name in profiles():
        fn, spec = get_profile(name)
        if fn.__doc__:
            fmt.print(fn.__doc__)
        else:
            print(f' ## {name} - lacks docstring\n')


def delete_current(sys):
    delete(sys.get_constant(SystemConstant.DB_URI), sys)


def delete(uri, sys):
    try:
        if database_exists(uri):
            log.debug(f'Deleting database at {uri}')
            uri_parts = urisplit(uri)
            if uri_parts.scheme == SQLITE:
                path = clean_path(uri_parts.path)
                log.warning(f'Deleting {path}')
                unlink(path)
            elif uri_parts.scheme == POSTGRESQL:
                drop_database(uri)
            else:
                raise Exception(f'Unsupported URI {uri}')
        else:
            log.warning(f'No database at {uri} (so not deleting)')
    finally:
        sys.delete_constant(SystemConstant.DB_URI)
        sys.delete_constant(SystemConstant.DB_VERSION)


def delete_and_check(uri, force, sys):
    if force: delete(uri, sys)
    if database_exists(uri):
        raise Exception(f'A schema exists at {uri} (use {mm(FORCE)}?)')


def write(uri, profile, sys, base):
    fn, spec = get_profile(profile)
    log.info(f'Loading profile {profile}')
    db = sys.get_database(uri)  # writes schema automatically
    with db.session_context() as s:
        fn(s, base)
    log.info(f'Profile {profile} loaded successfully')
    sys.set_constant(SystemConstant.DB_URI, uri, force=True)
    sys.set_constant(SystemConstant.DB_VERSION, DB_VERSION, force=True)


def make_uri(base, scheme_or_uri):
    if scheme_or_uri == SQLITE:
        uri = sqlite_uri(base)
    elif scheme_or_uri == POSTGRESQL:
        uri = postgresql_uri()
    else:
        uri = scheme_or_uri
    log.debug(f'Using URI {uri}')
    return uri


def create(uri):
    log.debug(f'Creating database at {uri}')
    create_database(uri)


def load(sys, base, profile, scheme_or_uri, force=False):
    uri = make_uri(base, scheme_or_uri)
    delete_and_check(uri, force, sys)
    create(uri)
    write(uri, profile, sys, base)



from logging import getLogger

from .args import FIX
from ..lib.log import Record
from ..sql import ActivityTopicJournal, FileHash, FileScan, StatisticJournal, ActivityJournal

log = getLogger(__name__)


def validate(args, data):
    '''
## validate

    > ch2 validate

This is still in development.
    '''
    record = Record(log)
    check_all(record, data.db, fix=args[FIX])


def check_all(record, db, fix=False):
    # todo - add flags
    check_activity_diary_missing_files(record, db, fix=fix)
    check_activity_journal_groups(record, db, fix=fix)


def check_activity_diary_missing_files(record, db, fix=False):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            bad = False
            for topic_journal in s.query(ActivityTopicJournal). \
                    join(StatisticJournal, StatisticJournal.source_id == ActivityTopicJournal.id). \
                    join(FileHash). \
                    outerjoin(FileScan). \
                    filter(FileScan.path == None).all():
                bad = True
                record.warning(f'{ActivityTopicJournal.__table__} with file hash '
                               f'{topic_journal.file_hash.hash[:6]} has associated entries but no activity')
                if fix:
                    record.warning('Deleting entry')
                    s.delete(topic_journal)
            if not bad:
                record.info('No missing activities from activity topic data')


def check_activity_journal_groups(record, db, fix=False):
    with record.record_exceptions(catch=True):
        with db.session_context() as s:
            bad = False
            for topic_journal, activity_journal in s.query(ActivityTopicJournal, ActivityJournal). \
                    select_from(ActivityTopicJournal). \
                    join(ActivityJournal, ActivityJournal.file_hash_id == ActivityTopicJournal.file_hash_id). \
                    filter(ActivityJournal.activity_group_id != ActivityTopicJournal.activity_group_id).all():
                bad = True
                record.warning(f'{topic_journal.activity_group.name} / {activity_journal.activity_group.name} '
                               f'for same activity ({activity_journal.start})')
                if fix:
                    record.warning('Changing topic to match activity')
                    topic_journal.activity_group = activity_journal.activity_group
            if not bad:
                record.info('No conflicting activity groups for activities')




# add flag fix to fix things

# check everything needed for power estimates
# (including weight)

# check constants defined

# check configure was run (version)

# check whether permanent directory present

# check whether file system ok (space?)

# check whether all files are scanned

# check whether all activity topics entries have an activity for the file hash

# all activity diary entries have corresponding activity - fix group

# check for activity without GPS datya (lyn in 2018)

# check for multiple activiy topic journals (or whatever we have)



import re
from re import sub

UNDEF = object()

SPACE = '-'  # we don't use _ because sqlite uses that as a wildcard in 'like'
POW_M1 = '\u207b\u00b9'
POW_2 = '\u00b2'
MED_WINDOW = '60s'


class Units:

    BPM = 'bpm'
    D = 'd'
    DEG = 'deg'
    FF = 'FF'
    H = 'h'
    J = 'J'
    KJ = 'kJ'
    KCAL = 'kCal'
    KG = 'kg'
    KM = 'km'
    KMH = f'kmh{POW_M1}'
    M = 'm'
    MS = f'ms{POW_M1}'
    PC = '%'
    RPM = 'rpm'
    S = 's'
    STEPS_UNITS = 'stp'
    W = 'W'


class TitlesBase:

    @staticmethod
    def _cov(name): return f'Cov {name}'  # coverage

    @staticmethod
    def _delta(name): return f'Δ {name}'

    @staticmethod
    def _avg(name): return f'Avg {name}'

    @staticmethod
    def _slash(name, units): return f'{name} / {units}'

    @staticmethod
    def _log(name): return f'Log {name}'

    @staticmethod
    def _sqr(name): return f'{name}{POW_2}'

    @staticmethod
    def _new(name): return f'New {name}'

    @staticmethod
    def _src(name): return f'Src {name}'

    @staticmethod
    def _lo(name): return f'Lo {name}'

    @staticmethod
    def _hi(name): return f'Hi {name}'

    @staticmethod
    def _s(name): return f'{name}s'

    @staticmethod
    def _med(name): return f'Med{MED_WINDOW} {name}'


class Titles(TitlesBase):

    AGE = 'Age'
    ACTIVE_DISTANCE = 'Active Distance'
    ACTIVE_SPEED = 'Active Speed'
    ACTIVE_TIME = 'Active Time'
    ACTIVITY = 'Activity'
    ACTIVITY_GROUP = 'Activity Group'
    AIR_SPEED = 'Air Speed'
    ALL = 'All'
    ANY_ALL = '% All'
    ALTITUDE = 'Altitude'
    ASPECT_RATIO = 'Aspect Ratio'
    BOOKMARK = 'Bookmark'
    CADENCE = 'Cadence'
    CALORIE_ESTIMATE = 'Calorie Estimate'
    CDA = 'CdA'
    CLIMB = 'Climb'
    CLIMB_ANY = 'Climb %'
    CLIMB_CATEGORY = 'Climb Category'
    CLIMB_DISTANCE = 'Climb Distance'
    CLIMB_ELEVATION = 'Climb Elevation'
    CLIMB_GRADIENT = 'Climb Gradient'
    CLIMB_TIME = 'Climb Time'
    CLIMB_POWER = 'Climb Power'
    COLOR = 'Color'
    COVERAGE = 'Coverage'
    CRR = 'Crr'
    CUMULATIVE_STEPS = 'Cumulative Steps'
    CUMULATIVE_STEPS_START = 'Cumulative Steps Start'
    CUMULATIVE_STEPS_FINISH = 'Cumulative Steps Finish'
    DAILY_STEPS = 'Daily Steps'
    DEFAULT = 'Default'
    DEFAULT_ANY = 'Default %'
    DELAYED_POWER = 'Delayed Power'
    DETRENDED_HEART_RATE = 'Detrended Heart Rate'
    DIRECTION = 'Direction'
    DISTANCE = 'Distance'
    EARNED_D = 'Earned %d'
    EARNED_D_ANY= 'Earned %'
    ELEVATION = 'Elevation'
    ENERGY = 'Energy'
    ENERGY_ESTIMATE = 'Energy Estimate'
    FATIGUE = 'Fatigue'
    FINISH = 'Finish'
    FITNESS = 'Fitness'
    FATIGUE_D = 'Fatigue %dd'
    FATIGUE_ANY = 'Fatigue %'
    ANY_FATIGUE_ANY = '% Fatigue %'
    FITNESS_D = 'Fitness %dd'
    FITNESS_ANY = 'Fitness %'
    ANY_FITNESS_ANY = '% Fitness %'
    FTHR = 'FTHR'
    GRADE = 'Grade'
    GROUP = 'Group'
    HEADING = 'Heading'
    HEART_RATE = 'Heart Rate'
    HR_IMPULSE_10 = 'HR Impulse / 10s'
    HR_ZONE = 'HR Zone'
    INDEX = 'Index'
    KIT_ADDED = 'Kit Added'
    KIT_RETIRED = 'Kit Retired'
    KIT_USED = 'Kit Used'
    LATITUDE = 'Latitude'
    LIFETIME = 'Lifetime'
    LOCAL_TIME = 'Date'
    LONGITUDE = 'Longitude'
    LOSS = 'Loss'
    MAX_MEAN_PE_M = 'Max Mean PE %dm'
    MAX_MEAN_PE_M_ANY = 'Max Mean PE %'
    MAX_MED_HR_M = 'Max Med HR %dm'
    MAX_MED_HR_M_ANY = 'Max Med HR %'
    MEAN_POWER_ESTIMATE = 'Mean Power Estimate'
    MED_KM_TIME = 'Med %dkm Time'
    MED_KM_TIME_ANY = 'Med % Time'
    MIN_KM_TIME = 'Min %dkm Time'
    MIN_KM_TIME_ANY = 'Min % Time'
    MIXED = 'Mixed'
    NAME = 'Name'
    PERCENT_IN_Z = 'Percent in Z%d'
    PERCENT_IN_Z_ANY = 'Percent in Z%'
    PLATEAU_D = 'Plateau %d'
    PLATEAU_D_ANY = 'Plateau %'
    POWER_ESTIMATE = 'Power Estimate'
    PREDICTED_HEART_RATE = 'Predicted Heart Rate'
    RESPONSES = 'Responses'
    RECOVERY_D = 'Recovery %d'
    RECOVERY_D_ANY = 'Recovery %'
    SEGMENT_TIME = 'Segment Time'
    SEGMENT_DISTANCE = 'Segment Distance'
    SEGMENT_HEART_RATE = 'Segment Heart Rate'
    SOURCE = 'Source'
    SPEED = 'Speed'
    SPHERICAL_MERCATOR_X = 'Spherical Mercator X'
    SPHERICAL_MERCATOR_Y = 'Spherical Mercator Y'
    START = 'Start'
    STEPS = 'Steps'
    RAW_ELEVATION = 'Raw Elevation'
    REST_HR = 'Rest HR'
    TIME = 'Time'
    TIME_IN_Z = 'Time in Z%d'
    TIME_IN_Z_ANY = 'Time in Z%'
    TIMESPAN_ID = 'Timespan ID'
    TOTAL_CLIMB = 'Total Climb'

    LO_REST_HR = TitlesBase._lo(REST_HR)
    HI_REST_HR = TitlesBase._hi(REST_HR)

    AIR_SPEED_2 = TitlesBase._sqr(AIR_SPEED)
    AVG_AIR_SPEED_2 = TitlesBase._avg(AIR_SPEED_2)
    DELTA_AIR_SPEED_2 = TitlesBase._delta(AIR_SPEED_2)

    SPEED_2 = TitlesBase._sqr(SPEED)
    AVG_SPEED_2 = TitlesBase._avg(SPEED_2)
    DELTA_SPEED_2 = TitlesBase._delta(SPEED_2)

    DELTA_TIME = TitlesBase._delta(TIME)
    DELTA_DISTANCE = TitlesBase._delta(DISTANCE)
    DELTA_ELEVATION = TitlesBase._delta(ELEVATION)
    DELTA_SPEED = TitlesBase._delta(SPEED)
    DELTA_ENERGY = TitlesBase._delta(ENERGY)

    ACTIVE_DISTANCE_KM = TitlesBase._slash(ACTIVE_DISTANCE, Units.KM)
    ACTIVE_TIME_H = TitlesBase._slash(ACTIVE_TIME, Units.H)
    ACTIVE_TIME_S = TitlesBase._slash(ACTIVE_TIME, Units.S)
    CADENCE_RPM = TitlesBase._slash(CADENCE, Units.RPM)
    CLIMB_MS = TitlesBase._slash(CLIMB, Units.MS)
    DISTANCE_KM = TitlesBase._slash(DISTANCE, Units.KM)
    ELEVATION_M = TitlesBase._slash(ELEVATION, Units.M)
    GRADE_PC = TitlesBase._slash(GRADE, Units.PC)
    HEART_RATE_BPM = TitlesBase._slash(HEART_RATE, Units.BPM)
    REST_HR_BPM = TitlesBase._slash(REST_HR, Units.BPM)
    SPEED_KMH = TitlesBase._slash(SPEED, Units.KMH)
    SPEED_MS = TitlesBase._slash(SPEED, Units.MS)
    POWER_ESTIMATE_W = TitlesBase._slash(POWER_ESTIMATE, Units.W)

    MED_SPEED_KMH = TitlesBase._med(SPEED_KMH)
    MED_HEART_RATE_BPM = TitlesBase._med(HEART_RATE_BPM)
    MED_HR_IMPULSE_10 = TitlesBase._med(HR_IMPULSE_10)
    MED_CADENCE_RPM = TitlesBase._med(CADENCE_RPM)
    MED_POWER_ESTIMATE_W = TitlesBase._med(POWER_ESTIMATE_W)

    POWER_HR = 'Power / HR'
    POWER_HR_LAG = 'Power / HR Lag'
    HR_DRIFT = 'HR Drift'
    LOG_HR_DRIFT = TitlesBase._log('HR Drift')
    IDLE_HR = 'Idle HR'
    WIND_SPEED = 'Wind Speed'
    WIND_HEADING = 'Wind Heading'


class Sports:

    SPORT_MAP = 'SportMap'
    SPORT_CYCLING = 'cycling'
    SPORT_RUNNING = 'running'
    SPORT_SWIMMING = 'swimming'
    SPORT_WALKING = 'walking'
    SPORT_GENERIC = 'generic'


class Summaries:

    MAX = 'max'
    MIN = 'min'
    CNT = 'cnt'
    AVG = 'avg'
    SUM = 'sum'
    MSR = 'msr'   # measure (separate table, less efficient)

    @staticmethod
    def join(*args): return ','.join(args)


def like(pattern, names):
    return list(_like(pattern, names))


def _like(pattern, names, test=True):
    matcher = re.compile(pattern.replace('%', '.+'))
    for name in names:
        if bool(matcher.match(name)) == test:
            yield name


def unlike(pattern, names):
    return list(_like(pattern, names, test=False))


def titles_for_names(title_pattern, names):
    title_left, title_right = title_pattern.split('%')
    name_left, name_right = simple_name(title_pattern).split('%')
    pattern = re.compile(f'{name_left}(.+){name_right}')
    for name in names:
        match = pattern.match(name)
        if match:
            yield title_left + match.group(1) + title_right


class NamesMeta(type):

    def __getattribute__(cls, item):
        value = type.__getattribute__(cls, item)
        if item[0].isalpha() and item[0].upper() == item[0]:
            return simple_name(value)
        elif len(item) > 1 and item[0] == '_' and item[1].isalpha():
            def wrapper(*args):
                return simple_name(value(*args))
            return wrapper
        else:
            return value


class Names(Titles, metaclass=NamesMeta): pass


T, N = Titles, Names


def simple_name(name, none=True, strip=True):
    # allows % and ? for LIKE and templates (although sqlite uses _ instead of ?)
    # also allows ':' so that we don't mess up composites
    from ch2.names import POW_2, POW_M1, SPACE
    if name is None and none:
        return None
    name = name.replace(POW_2, '2')
    name = name.replace(POW_M1, '')  # ms^-1 -> ms which is standard convention
    name = name.replace('Δ', 'd ')
    if strip: name = name.strip()
    name = name.lower()
    name = sub(r'\s+', SPACE, name)
    name = sub(r'[^a-z0-9%?:]', SPACE, name)
    name = sub(r'^(\d)', SPACE + r'\1', name)
    name = sub(SPACE + '+', SPACE, name)
    return name



from itertools import zip_longest
from json import loads
from logging import getLogger

from ..lib.utils import clean
from ..names import SPORT_CYCLING, SPORT_RUNNING, SPORT_SWIMMING, SPORT_WALKING, SPORT_GENERIC, SPORT_MAP
from ..sql import KitItem, ActivityGroup
from ..sql.tables.constant import ValidateError, Validate, Constant

log = getLogger(__name__)


class SportMap:

    SPORTS = [clean(s) for s in (SPORT_CYCLING, SPORT_RUNNING, SPORT_SWIMMING, SPORT_WALKING, SPORT_GENERIC)]

    def __init__(self, json):
        '''
        the json representation is a list of string triplets: [(sport, kit, group), ...]
        where sport matches the sport in the FIT file, kit matches any kit specified by the user on upload,
        and group identifies the activity group that will include the activity.
        first match 'wins' and None for sport or kit is a wildcard.
        None for group is equivalent to ALL.
        there is an implicit final entry of (None, None, None) which matches anything to ALL.
        '''
        if isinstance(json, str):
            json = loads(json)
        self.__json = json

    def add(self, s, error=True):
        self.validate(s, error=error)
        if s.query(Constant).filter(Constant.name == SPORT_MAP).count():
            raise Exception(f'Constant {SPORT_MAP} already exists')
        constant = Constant(name=SPORT_MAP, )


    class Validator(Validate):

        def __init__(self, error=True):
            self.__error = error

        def validate(self, s, constant, sjournal):
            try:
                json = loads(sjournal.value)
            except Exception as e:
                raise ValidateError('Could not unpack JSON value for %s from "%s": %s' %
                                    (sjournal.value, constant.name, e))
            SportMap.validate_json(s, json, error=self.__error)

    def validate(self, s, error=True):
        self.validate_json(s, self.__json, error=error)

    @classmethod
    def validate_json(cls, s, json, error=True):
        try:
            for entries in json:
                for entry, validator in zip_longest(entries,
                                                    (cls.validate_sport, cls.validate_kit, cls.validate_group)):
                    try:
                        validator(s, entry)
                    except ValidateError as e:
                        if error:
                            log.error(e)
                            raise
                        else:
                            log.warning(e)
        except ValidateError:
            raise
        except Exception as e:
            log.error(f'Cannot validate {json}: {e}')
            # this is always an error - we need to be able to parse the thing
            raise

    @classmethod
    def validate_sport(cls, s, sport):
        if clean(sport) not in cls.SPORTS:
            raise ValidateError(f'Unknown sport {sport}')

    @classmethod
    def validate_kit(cls, s, kit):
        if not s.query(KitItem).filter(KitItem.name == kit).count():
            raise ValidateError(f'Unknown kit {kit}')

    @classmethod
    def validate_group(cls, s, group):
        if not s.query(ActivityGroup).filter(ActivityGroup.name == group).count():
            raise ValidateError(f'Unknown group {group}')


from collections import defaultdict
from os.path import splitext


class ContentType:

    CONTENT_TYPE = defaultdict(lambda: 'text/plain', {
        'js': 'text/javascript',
        'html': 'text/html',
        'css': 'text/css',
        'png': 'image/png'
    })

    def set_content_type(self, response, name):
        ext = splitext(name)[1].lower()
        if ext:
            ext = ext[1:]
        response.content_type = self.CONTENT_TYPE[ext]




from logging import getLogger

from sqlalchemy import or_

from .utils import Displayer
from ..calculate.summary import SummaryCalculator
from ...diary.database import interval_column
from ...diary.model import text, from_field, optional_text, value, trim_no_stats
from ...sql import DiaryTopicJournal, DiaryTopic, StatisticJournal, Interval

log = getLogger(__name__)


class DiaryDisplayer(Displayer):

    @optional_text('Diary')
    def _read_date(self, s, date):
        journal = DiaryTopicJournal.get_or_add(s, date)
        for topic in s.query(DiaryTopic).filter(DiaryTopic.parent == None,
                                                or_(DiaryTopic.start <= date, DiaryTopic.start == None),
                                                or_(DiaryTopic.finish >= date, DiaryTopic.finish == None)). \
                order_by(DiaryTopic.sort).all():
            if topic.schedule.at_location(date):
                yield list(self.__read_date_diary_topic(s, date, journal.cache(s), topic))

    def __read_date_diary_topic(self, s, date, cache, topic):
        yield text(topic.title)
        if topic.description: yield text(topic.description)
        log.debug(f'topic id {topic.id}; fields {topic.fields}')
        for field in topic.fields:
            if field.schedule.at_location(date):
                yield from_field(field, cache[field])
        for child in topic.children:
            if child.schedule.at_location(date):
                content = list(self.__read_date_diary_topic(s, date, cache, child))
                if content:
                    # single entries are just text fields
                    if len(content) == 1:
                        yield content[0]
                    else:
                        yield content

    @optional_text('Diary')
    @trim_no_stats
    def _read_schedule(self, s, date, schedule):
        interval = s.query(Interval). \
            filter(Interval.schedule == schedule,
                   Interval.start == date,
                   Interval.activity_group == None).one_or_none()
        if interval:
            for topic in s.query(DiaryTopic).filter(DiaryTopic.parent == None,
                                                    or_(DiaryTopic.start < interval.finish,
                                                        DiaryTopic.start == None),
                                                    or_(DiaryTopic.finish >= date,
                                                        DiaryTopic.finish == None)). \
                    order_by(DiaryTopic.sort).all():
                yield list(self.__read_interval_topic(s, interval, topic))

    def __read_interval_topic(self, s, interval, topic):
        yield text(topic.title)
        if topic.description: yield text(topic.description)
        for field in topic.fields:
            column = list(interval_column(s, interval, field.statistic_name.name, SummaryCalculator))
            if column: yield column
        for child in topic.children:
            if (child.start is None or child.start < interval.finish) and \
                    (child.finish is None or child.finish > interval.start):
                content = list(self.__read_interval_topic(s, interval, child))
                if content: yield content




import datetime as dt
from abc import ABC, abstractmethod
from json import dumps, loads
from logging import getLogger

from sqlalchemy import Column, Integer, ForeignKey, Boolean, desc
from sqlalchemy.orm import relationship

from .source import Source, SourceType
from .statistic import STATISTIC_JOURNAL_CLASSES, StatisticJournal
from ..types import Cls, Json, lookup_cls, QualifiedName
from ...lib.date import local_date_to_time, format_time, to_time
from ...lib.log import log_current_exception

log = getLogger(__name__)


class Constant(Source):

    __tablename__ = 'constant'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    # this could be the statistic_name or it could contain more info related to constraint
    name = Column(QualifiedName, nullable=False, index=True, unique=True)
    # todo - this ondelete cascade could cause problems with orphaned sources
    # don't think it's needed anyway, since statistic_name entries are not deleted?
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
    statistic_name = relationship('StatisticName')
    single = Column(Boolean, nullable=False, server_default='0')
    validate_cls = Column(Cls)
    validate_args = Column(Json, nullable=False, server_default=dumps(()))
    validate_kargs = Column(Json, nullable=False, server_default=dumps({}))

    def validate(self, s, sjournal):
        if self.validate_cls:
            if self.validate_args is None or self.validate_kargs is None:
                raise Exception('Missing args or kargs for %s' % self)
            validate = self.validate_cls(*self.validate_args, **self.validate_kargs)
            validate.validate(s, self, sjournal)

    def add_value(self, s, value, time=None, date=None):
        from ..utils import add
        if time and date:
            raise Exception('Specify one or none of time and date for %s' % self)
        if not time and not date:
            time = 0.0  # important this is a float and not an int (or it would be an erroneous date)
        if date:
            time = local_date_to_time(date)
        if time and self.single:
            raise Exception('%s was given time %s but is not time-variable' % (self, format_time(to_time(time))))
        sjournal = STATISTIC_JOURNAL_CLASSES[self.statistic_name.statistic_journal_type](
            statistic_name=self.statistic_name, source=self, value=value, time=time)
        self.validate(s, sjournal)
        return add(s, sjournal)

    def at(self, s, time=None, date=None):
        if time and date:
            raise Exception('Specify one or none of time and date for %s' % self)
        if date:
            time = local_date_to_time(date)
        if not time:
            time = dt.datetime.now(tz=dt.timezone.utc)
        return s.query(StatisticJournal). \
            filter(StatisticJournal.statistic_name == self.statistic_name,
                   StatisticJournal.time <= time,
                   StatisticJournal.source == self). \
            order_by(desc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def from_name(cls, s, name, none=False):
        constant = s.query(Constant).filter(Constant.name == name).one_or_none()
        if constant is None and not none:
            raise Exception('Could not find Constant for %s' % name)
        return constant

    @classmethod
    def get_single(cls, s, name):
        try:
            constant = Constant.from_name(s, name)
            if not constant.single:
                raise Exception(f'Constant {name} is not single')
            value = constant.at(s).value
            log.debug(f'{name} is {value}')
            return value
        except Exception as e:
            log_current_exception(traceback=False)
            raise Exception(f'{name} is not configured')

    @property
    def short_name(self):
        if ':' in self.name:
            return self.name.split(':')[0]
        else:
            return self.name

    __mapper_args__ = {
        'polymorphic_identity': SourceType.CONSTANT
    }

    def __str__(self):
        return 'Constant "%s"' % self.name

    def time_range(self, s):
        return None, None


class Validate(ABC):

    @abstractmethod
    def validate(self, s, constant, sjournal):
        raise NotImplementedError()


class ValidateError(Exception): pass


class ValidateNamedTuple(Validate):

    def __init__(self, tuple_cls=None):
        if not tuple_cls:
            raise Exception('Add tuple_cls to validator_kargs')
        self.__tuple_cls = lookup_cls(tuple_cls)

    def validate(self, s, constant, sjournal):
        try:
            value = loads(sjournal.value)
        except Exception as e:
            raise ValidateError('Could not unpack JSON value for %s from "%s": %s' %
                                (sjournal.value, constant.name, e))
        try:
            self.__tuple_cls(**value)
        except Exception as e:
            raise ValidateError('Could not create %s from "%s" for "%s": %s' %
                                (self.__tuple_cls, value, constant.name, e))





import datetime as dt
from abc import abstractmethod
from logging import getLogger

from ..pipeline import BasePipeline
from ...lib import local_date_to_time, log_current_exception
from ...sql import ActivityGroup, ActivityJournal

log = getLogger(__name__)


class Displayer(BasePipeline):

    def read(self, s, date, schedule=None):
        try:
            if schedule:
                yield from self._read_schedule(s, date, schedule)
            else:
                yield from self._read_date(s, date)
        except Exception as e:
            log_current_exception(e)

    @abstractmethod
    def _read_schedule(self, s, date, schedule):
        raise NotImplementedError(self.__class__.__name__)
        yield

    @abstractmethod
    def _read_date(self, s, date):
        raise NotImplementedError(self.__class__.__name__)
        yield


class ActivityJournalDelegate(BasePipeline):

    def __init__(self, interpolate=False):
        self.interpolate = interpolate

    @abstractmethod
    def read_journal_date(self, s, ajournal, date):
        raise NotImplementedError(self.__class__.__name__)
        yield

    @abstractmethod
    def read_interval(self, s, interval):
        raise NotImplementedError(self.__class__.__name__)
        yield


class JournalDiary(Displayer):

    def _read_date(self, s, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for ajournal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield from self._read_journal_date(s, ajournal, date)

    @abstractmethod
    def _read_journal_date(self, s, ajournal, date):
        raise NotImplementedError(self.__class__)
        yield



import datetime as dt
from logging import getLogger

from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint, desc
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType, GroupedSource
from ..support import Base
from ..types import Time, Sort, ShortCls, NullText, Name, name_and_title
from ...lib.date import format_time, local_date_to_time, local_time_to_time
from ...lib.utils import timing
from ...names import Titles, UNDEF, simple_name

log = getLogger(__name__)


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Name, nullable=False, index=True, unique=True)
    title = Column(Text, nullable=False)
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False)

    def __init__(self, **kargs):
        super().__init__(**name_and_title(kargs))

    def __str__(self):
        return self.name

    @classmethod
    def from_name(cls, s, name, none=False):
        if not name:
            return None
        elif isinstance(name, ActivityGroup):
            return name  # allow callers to already have an instance
        else:
            instance = s.query(ActivityGroup).filter(ActivityGroup.name.ilike(name)).one_or_none()
            if instance is None and not none:
                raise Exception(f'No activity group defined for {name}')
            return instance


class ActivityJournal(GroupedSource):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False, index=True, unique=True)
    file_hash = relationship('FileHash', backref=backref('activity_journal', uselist=False))
    start = Column(Time, nullable=False, index=True, unique=True)
    finish = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    def get_all_qname(self, s, qname, limit=True):
        direct = self.get_qname(s, qname, limit=limit)
        indirect = self.get_activity_topic_journal(s).get_qname(s, qname, limit=limit)
        log.debug(f'{len(direct)} {len(indirect)}')
        return direct + indirect

    def get_activity_topic_journal(self, s):
        from .. import ActivityTopicJournal, FileHash
        with timing('get_activity_topic_journal'):
            return s.query(ActivityTopicJournal). \
                join(FileHash). \
                join(ActivityJournal). \
                filter(ActivityJournal.id == self.id).one()

    @classmethod
    def at(cls, s, local_time_or_date, activity_group=None):
        try:
            activity_journal = cls.at_local_time(s, local_time_or_date, activity_group=activity_group)
        except:
            activity_journal = cls.at_date(s, local_time_or_date, activity_group=activity_group)
        if activity_group and activity_journal.activity_group != ActivityGroup.from_name(s, activity_group):
            raise Exception(f'Activity journal from {local_time_or_date} '
                            f'does not match activity group {activity_group}')
        log.debug(f'Resolved {local_time_or_date} / {activity_group} to {activity_journal}')
        return activity_journal

    @classmethod
    def at_date(cls, s, date, activity_group=None):
        day = local_date_to_time(date)
        journals = s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                                   ActivityJournal.start < day + dt.timedelta(days=1)).all()
        if activity_group:
            activity_group = ActivityGroup.from_name(s, activity_group)
            journals = [journal for journal in journals if journal.activity_group == activity_group]
        if not journals:
            raise Exception(f'No activity journal found at {date} for activity group {activity_group}')
        elif len(journals) > 1:
            raise Exception(f'Multiple activity journals found at {date} for activity group {activity_group}')
        else:
            return journals[0]

    @classmethod
    def at_local_time(cls, s, local_time, activity_group=None):
        time = local_time_to_time(local_time)
        q = s.query(ActivityJournal).filter(ActivityJournal.start == time)
        if activity_group:
            activity_group = ActivityGroup.from_name(s, activity_group)
            q = q.filter(ActivityJournal.activity_group_id == activity_group.id)
        return q.one()

    @classmethod
    def before_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start < time). \
            order_by(desc(ActivityJournal.start)). \
            limit(1).one_or_none()

    @classmethod
    def after_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start >= (time + dt.timedelta(days=1))). \
            order_by(ActivityJournal.start). \
            limit(1).one_or_none()

    @classmethod
    def number_of_activities(cls, s):
        return s.query(ActivityJournal).count()

    def __lt__(self, other):
        # allows sorting in displays
        if isinstance(other, ActivityJournal):
            return self.start < other.start
        else:
            return False


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))


class ActivityBookmark(Base):

    __tablename__ = 'activity_bookmark'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
    activity_journal = relationship('ActivityJournal')
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    constraint = Column(NullText, index=True)
    UniqueConstraint(activity_journal_id, start, finish, owner, constraint)

    def __str__(self):
        return 'ActivityBookmark from %s - %s' % (format_time(self.start), format_time(self.finish))





from logging import getLogger

from .utils import Displayer
from ..calculate import SummaryCalculator
from ..calculate.monitor import MonitorCalculator
from ...names import N
from ...diary.database import summary_column, interval_column
from ...diary.model import value, optional_text
from ...sql import Interval
from ...sql.tables.statistic import StatisticJournal, StatisticName

log = getLogger(__name__)


class MonitorDisplayer(Displayer):

    @optional_text('Monitor')
    def _read_date(self, s, date):
        for field in self.__read_fields(s, date, N.DAILY_STEPS, N.REST_HR):
            yield value(field.statistic_name.title, field.value,
                        units=field.statistic_name.units, measures=field.measures_as_model(date))

    @staticmethod
    def __read_fields(s, date, *names):
        for name in names:
            journal = StatisticJournal.at_date(s, date, name, MonitorCalculator, None)
            if journal:
                yield journal

    @optional_text('Monitor')
    def _read_schedule(self, s, date, schedule):
        interval = s.query(Interval). \
            filter(Interval.schedule == schedule,
                   Interval.start == date,
                   Interval.activity_group == None).one_or_none()
        for name in (N.DAILY_STEPS, N.REST_HR):
            column = list(interval_column(s, interval, name, SummaryCalculator))
            if column: yield column




from .nearby import fmt_nearby, nearby_any_time
from ..utils import ActivityJournalDelegate
from ....diary.model import text, link, optional_text, COMPARE_LINKS
from ....lib import time_to_local_time
from ....lib.date import format_date, local_date_to_time
from ....sql import ActivityJournal


class JupyterDelegate(ActivityJournalDelegate):

    @optional_text('Jupyter', tag='jupyter-activity')
    def read_journal_date(self, s, ajournal, date):
        links = [link('None', db=(time_to_local_time(ajournal.start), None, ajournal.activity_group.name))] + \
                [link(fmt_nearby(ajournal2, nb),
                      db=(time_to_local_time(ajournal.start), time_to_local_time(ajournal2.start), ajournal.activity_group.name))
                 for ajournal2, nb in nearby_any_time(s, ajournal)]
        yield [text('Compare to', tag=COMPARE_LINKS)] + links
        yield link('All Similar', db=(time_to_local_time(ajournal.start), ajournal.activity_group.name))

    @optional_text('Jupyter', tag='jupyter-activity')
    def read_interval(self, s, interval):
        if s.query(ActivityJournal). \
            filter(ActivityJournal.start > local_date_to_time(interval.start),
                   ActivityJournal.finish < local_date_to_time(interval.finish),
                   ActivityJournal.activity_group_id == interval.activity_group_id).count():
            yield link('Group Activities', db=(format_date(interval.start), format_date(interval.finish),
                                               interval.activity_group.name))





from sqlalchemy import or_, desc
from sqlalchemy.orm import aliased

from ..utils import ActivityJournalDelegate
from ....diary.model import text, link, optional_text
from ....lib.date import to_time, time_to_local_time
from ....sql import ActivityJournal, ActivitySimilarity
from ....sql.support import greatest

NEARBY_LINKS = 'nearby-links'


def fmt_nearby(aj, nb):
    return to_time(aj.start).strftime('%y-%m-%d') + ' %d%%' % (nb.similarity * 100)


class NearbyDelegate(ActivityJournalDelegate):

    @optional_text('Nearby', tag='nearby')
    def read_journal_date(self, s, ajournal, date):
        yield from self.__read_nearby(s, ajournal)

    def __read_nearby(self, s, ajournal):
        for title, callback in (('Any Time', nearby_any_time), ('Earlier', nearby_earlier)):
            links = [link(fmt_nearby(aj, nb), db=(time_to_local_time(aj.start),))
                     for (aj, nb) in callback(s, ajournal)]
            if links:
                yield [text(title, tag=NEARBY_LINKS)] + links

    def read_interval(self, s, interval):
        return
        yield


def nearby_earlier(s, ajournal, threshold=0.3):
    aj_lo = aliased(ActivityJournal)
    aj_hi = aliased(ActivityJournal)
    q = s.query(ActivitySimilarity). \
        join(aj_hi, ActivitySimilarity.activity_journal_hi_id == aj_hi.id). \
        join(aj_lo, ActivitySimilarity.activity_journal_lo_id == aj_lo.id). \
        filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                   ActivitySimilarity.activity_journal_lo_id == ajournal.id),
               aj_lo.activity_group_id == aj_hi.activity_group_id,
               or_(aj_hi.id == ajournal.id, aj_hi.start < ajournal.start),
               or_(aj_hi.id == ajournal.id, aj_lo.start < ajournal.start),
               ActivitySimilarity.similarity > threshold). \
        order_by(desc(greatest(aj_lo.start, aj_hi.start)))
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]


def nearby_any_time(s, ajournal, threshold=0.3):
    aj_lo = aliased(ActivityJournal)
    aj_hi = aliased(ActivityJournal)
    q = s.query(ActivitySimilarity). \
        join(aj_hi, ActivitySimilarity.activity_journal_hi_id == aj_hi.id). \
        join(aj_lo, ActivitySimilarity.activity_journal_lo_id == aj_lo.id). \
        filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                   ActivitySimilarity.activity_journal_lo_id == ajournal.id),
               aj_lo.activity_group_id == aj_hi.activity_group_id,
               ActivitySimilarity.similarity > threshold). \
        order_by(desc(ActivitySimilarity.similarity))
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]




from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np

from .utils import MultiProcCalculator, ActivityGroupCalculatorMixin, DataFrameCalculatorMixin
from ..pipeline import OwnerInMixin
from ...data import Statistics
from ...data.frame import valid
from ...data.impulse import hr_zone, impulse_10
from ...names import N, Titles, SPACE
from ...sql import Constant, StatisticJournalFloat

log = getLogger(__name__)

# constraint comes from constant
HRImpulse = namedtuple('HRImpulse', 'title, gamma, zero, one, max_secs')


class ImpulseCalculator(OwnerInMixin, ActivityGroupCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, prefix=None, impulse_constant=None, **kargs):
        self.impulse_constant_ref = self._assert('impulse_constant', impulse_constant)
        self.prefix = self._assert('prefix', prefix)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        self.impulse_constant = Constant.from_name(s, self.impulse_constant_ref)
        self.impulse = HRImpulse(**loads(self.impulse_constant.at(s).value))
        log.debug('%s: %s' % (self.impulse_constant, self.impulse))

    def _read_dataframe(self, s, ajournal):
        try:
            heart_rate_df = Statistics(s, activity_journal=ajournal). \
                by_name(self.owner_in, N.HEART_RATE).df
            fthr_df = Statistics(s).by_name(Constant, N.FTHR).df
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise
        if fthr_df.empty:
            raise Exception(f'No {N.FTHR} defined for {ajournal.activity_group}')
        return heart_rate_df, fthr_df

    def _calculate_stats(self, s, ajournal, data):
        heart_rate_df, fthr_df = data
        hr_zone(heart_rate_df, fthr_df)
        impulse_df = impulse_10(heart_rate_df, self.impulse)
        # join so that we can iterate over values in time order
        stats = impulse_df.join(heart_rate_df, how='outer')
        return stats

    def _copy_results(self, s, ajournal, loader, stats):
        hr_description = 'The SHRIMP HR zone.'
        impulse_description = 'The SHRIMP HR impulse over 10 seconds.'
        title = self.impulse.title
        name_group = self.prefix + SPACE + self.impulse_constant.short_name  # drop activity group as present elsewhere
        for time, row in stats.iterrows():
            if N.HR_ZONE in row and valid(row[N.HR_ZONE]):
                loader.add(Titles.HR_ZONE, None, None, ajournal, row[N.HR_ZONE], time,
                           StatisticJournalFloat, description=hr_description)
            if N.HR_IMPULSE_10 in row and valid(row[N.HR_IMPULSE_10]):
                loader.add(name_group, None, None, ajournal, row[N.HR_IMPULSE_10], time,
                           StatisticJournalFloat, description=impulse_description, title=title)
        # if there are no values, add a single 1 so we don't re-process
        if not loader:
            loader.add(Titles.HR_ZONE, None, None, ajournal, 1, ajournal.start,
                       StatisticJournalFloat, description=hr_description)




from abc import abstractmethod
from enum import IntEnum
from logging import getLogger
from operator import or_

from sqlalchemy import ForeignKey, Column, Integer, func, UniqueConstraint
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql.functions import count

from ..support import Base
from ..types import OpenSched, Date, ShortCls, short_cls
from ..utils import add
from ...global_ import global_data
from ...lib.date import to_time, time_to_local_date, max_time, min_time, extend_range
from ...lib.utils import timing, grouper
from ...names import UNDEF

log = getLogger(__name__)


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    DIARY_TOPIC = 3
    CONSTANT = 4
    MONITOR = 5
    SEGMENT = 6
    COMPOSITE = 7
    DUMMY = 8
    ITEM = 9
    MODEL = 10
    ACTIVITY_TOPIC = 11


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast deletes of subtypes
    activity_group_id = Column(Integer, ForeignKey('activity_group.id', ondelete='cascade'), nullable=True)
    activity_group = relationship('ActivityGroup')

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @abstractmethod
    def time_range(self, s):
        raise NotImplementedError('time_range for %s' % self)

    @classmethod
    def before_flush(cls, s):
        cls.__record_dirty_intervals(s)

    @classmethod
    def __record_dirty_intervals(cls, s):
        from .statistic import StatisticJournal, StatisticJournalText
        # sessions are generally restricted to one time region, so we'll bracket that rather than list all times
        start, finish = None, None
        # all sources except intervals that are being deleted (need to catch on cascade to statistics)
        for instance in s.deleted:
            if isinstance(instance, Source) and not isinstance(instance, Interval):
                a, b = instance.time_range(s)
                start, finish = min_time(a, start), max_time(b, finish)
        # all modified statistics
        for instance in s.dirty:
            # ignore constants as time 0
            # ignore textual values (don't have useful stats and commonly modified fields)
            if isinstance(instance, StatisticJournal):
                if s.is_modified(instance) and instance.time and not isinstance(instance, StatisticJournalText):
                    start, finish = extend_range(start, finish, instance.time)
                elif instance.statistic_name is None or instance.source is None:
                    log.warning(f'Loading with incomplete name/source data (bad dirty logic) '
                                f'(statistic_name_id {instance.statistic_name_id})')
        # all new statistics that aren't associated with intervals and have non-null data
        # (avoid triggering on empty diary entries)
        for instance in s.new:
            # ignore constants as time 0
            # ignore textual values
            if isinstance(instance, StatisticJournal):
                if not isinstance(instance.source, Interval) and not isinstance(instance.source, Dummy) \
                        and not isinstance(instance.source, StatisticJournalText):
                    start, finish = extend_range(start, finish, instance.time)
                elif instance.statistic_name is None or instance.source is None:
                    log.warning(f'Loading with incomplete name/source data (bad dirty logic) '
                                f'(statistic_name_id {instance.statistic_name_id})')
        if start is not None:
            Interval.record_dirty_times(s, start, finish)

    @classmethod
    def from_id(cls, s, id):
        return s.query(Source).filter(Source.id == id).one()

    def long_str(self):
        return str(self)

    def get_qname(self, s, qname, limit=True):
        from .. import StatisticJournal, StatisticName, ActivityGroup
        owner, name, group = StatisticName.parse(qname, default_activity_group=UNDEF)
        q = s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name.ilike(name),
                   StatisticJournal.source_id == self.id)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if group is not UNDEF:
            q = q.join(ActivityGroup).filter(ActivityGroup.name == group)
        if limit:
            q = q.filter(or_(StatisticJournal.serial.is_(None), StatisticJournal.serial == 0))
        log.debug(q)
        with timing(f'get_qname {self.id} {owner} {name} {group}'):
            return q.all()



@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    from .. import StatisticJournal
    StatisticJournal.before_flush(session)
    Source.before_flush(session)


class GroupedSource(Source):

    __abstract__ = True

    def __init__(self, **kargs):
        if 'activity_group' not in kargs: raise Exception(f'{self.__class__.__name__} requires activity group')
        super().__init__(**kargs)

    def long_str(self):
        return super().long_str() + f' ({self.activity_group.name})'


class UngroupedSource(Source):

    __abstract__ = True

    def __init__(self, **kargs):
        if 'activity_group' in kargs: raise Exception(f'{self.__class__.__name__} has no activity group')
        super().__init__(**kargs)


class NoStatistics(Exception):
    pass


class Interval(Source):

    __tablename__ = 'interval'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False, index=True)
    owner = Column(ShortCls, nullable=False)
    # these are for the schedule - finish is redundant (start is not because of timezone issues)
    start = Column(Date, nullable=False, index=True)
    finish = Column(Date, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __str__(self):
        owner = self.owner if isinstance(self.owner, str) else short_cls(self.owner)
        return 'Interval "%s from %s" (owner %s)' % (self.schedule, self.start, owner)

    @classmethod
    def _raw_statistics_time_range(cls, s, statistics_owner=None):
        '''
        The time range over which statistics exist (optionally restricted by owner),
        ignoring constants at "time zero".  This is the first to the last time for
        any statistics - it pays no attention to gaps.
        '''
        from .statistic import StatisticJournal, StatisticName
        q = s.query(func.min(StatisticJournal.time), func.max(StatisticJournal.time)). \
            filter(StatisticJournal.time > to_time(2 * 24 * 60 * 60.0))
        if statistics_owner:
            q = q.join(StatisticName).filter(StatisticName.owner == statistics_owner)
        start, finish = q.one()   # skip entire first day because tz
        if start and finish:
            return start, finish
        else:
            raise NoStatistics('No statistics are currently defined')

    @classmethod
    def missing_dates(cls, s, expected, schedule, interval_owner, statistic_owner=None, start=None, finish=None):
        '''
        Previous approach was way too complicated and not thread-safe.  Instead, just enumerate intervals and test.
        '''
        stats_start_time, stats_finish_time = cls._raw_statistics_time_range(s, statistic_owner)
        stats_start = time_to_local_date(stats_start_time)
        stats_finish = time_to_local_date(stats_finish_time)
        log.debug('Statistics (in general) exist %s - %s' % (stats_start, stats_finish))
        start = schedule.start_of_frame(start if start else stats_start)
        finish = finish if finish else schedule.next_frame(stats_finish)
        while start < finish:
            next = schedule.next_frame(start)
            existing = s.query(Interval). \
                filter(Interval.start == start,
                       Interval.schedule == schedule,
                       Interval.owner == interval_owner).count()
            if existing != expected:
                yield start, next
            start = next

    @classmethod
    def dirty_all(cls, s):
        log.warning('Dirtying all Intervals')
        global_data().sys.record_dirty_intervals(interval.id for interval in s.query(Interval).all())

    @classmethod
    def record_dirty_times(cls, s, start, finish, owner=None):
        '''
        Record dirty intervals that include data in the given TIME range,
        '''
        start, finish = time_to_local_date(start), time_to_local_date(finish)
        q = s.query(Interval.id).filter(Interval.start <= finish, Interval.finish > start)
        if owner:
            q = q.filter(Interval.owner == owner)
        # do not mark in-place because we can get deadlock transactions.
        # instead, save in sys and update later
        global_data().sys.record_dirty_intervals(row[0] for row in q.all())

    @classmethod
    def clean(cls, s):
        data = global_data()
        dirty_intervals = data.sys.get_dirty_intervals()
        if dirty_intervals:
            log.debug('Cleaning dirty intervals')
            dirty_ids = set(d.interval_id for d in dirty_intervals)
            log.warning(f'Up to {len(dirty_ids)} intervals to delete')
            for ids in grouper(dirty_ids, 900):  # avoid sqlite limit
                s.query(Source).filter(Source.id.in_(ids)).delete(synchronize_session=False)
            s.commit()
            data.sys.delete_dirty_intervals(dirty_intervals)
            log.debug('Intervals clean')


class Dummy(UngroupedSource):

    __tablename__ = 'dummy_source'

    DUMMY = 'Dummy'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.DUMMY
    }

    def time_range(self, s):
        return None, None

    @classmethod
    def singletons(cls, s):
        from .. import StatisticName
        source = s.query(Dummy).one()
        name = s.query(StatisticName).filter(StatisticName.owner == source).one()
        return source, name


class CompositeComponent(Base):

    __tablename__ = 'composite_component'

    id = Column(Integer, primary_key=True)
    input_source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                             nullable=False, index=True)
    input_source = relationship('Source', foreign_keys=[input_source_id])
    output_source_id = Column(Integer, ForeignKey('composite_source.id', ondelete='cascade'),
                              nullable=False)  # no need for index - see unique below
    output_source = relationship('Composite', foreign_keys=[output_source_id])
    UniqueConstraint(output_source_id, input_source_id)  # ordered so output is a useful index


class Composite(Source):

    __tablename__ = 'composite_source'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    components = relationship('Source', secondary='composite_component')
    n_components = Column(Integer, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.COMPOSITE
    }

    def time_range(self, s):
        return None, None

    @classmethod
    def clean(cls, s):
        log.debug('Searching for invalid composites')
        # see test_recursive
        q_input_counts = s.query(Composite.id,
                                 count(CompositeComponent.input_source_id).label('count')). \
            outerjoin(CompositeComponent, CompositeComponent.output_source_id == Composite.id). \
            group_by(Composite.id).cte()
        q_bad_nodes = s.query(Composite.id). \
            join(q_input_counts, q_input_counts.c.id == Composite.id). \
            filter(Composite.n_components != q_input_counts.c.count)
        q_count = s.query(count(Composite.id)).filter(Composite.id.in_(q_bad_nodes))
        log.debug(q_count)
        if q_count.scalar():
            log.warning('Need to clean expired composite sources (may take some time)')
            q_bad_nodes = q_bad_nodes.cte(recursive=True)
            q_all_nodes = q_bad_nodes. \
                union_all(s.query(Composite.id).
                          join(CompositeComponent,
                               CompositeComponent.output_source_id == Composite.id).
                          join(q_bad_nodes,
                               CompositeComponent.input_source_id == q_bad_nodes.c.id)).select()
            log.debug(f'Executing {q_all_nodes}')
            s.flush()
            with timing('GC of composite sources'):
                s.query(Source).filter(Source.id.in_(q_all_nodes)).delete(synchronize_session=False)

    @classmethod
    def create(cls, s, *sources):
        composite = add(s, Composite(n_components=0))
        for source in sources:
            add(s, CompositeComponent(input_source=source, output_source=composite))
            composite.n_components += 1
        return composite

    def long_str(self):
        interesting = [component for component in self.components if not isinstance(component, Composite)]
        msg = f'Composite (id {self.id}, {self.n_components} comp)'
        if self.activity_group: msg += f' ({self.activity_group.name})'
        if interesting:
            msg += ' [' + ', '.join(source.long_str() for source in interesting) + ']'
        return msg




from logging import getLogger

import numpy as np
import pandas as pd
from sqlalchemy import asc, desc, distinct
from sqlalchemy.orm import aliased

from ch2.data import read_query
from ..data import session, present
from ..lib import local_date_to_time, to_date, time_to_local_time
from ..lib.date import YMD, format_seconds
from ..lib.log import log_current_exception
from ..lib.utils import timing
from ..names import Names as N, like, MED_WINDOW, SPACE
from ..sql import StatisticName, ActivityGroup, StatisticJournal, ActivityTimespan, ActivityJournal, Source, \
    ActivityTopic
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ..sql.types import short_cls

log = getLogger(__name__)


class Statistics:

    def __init__(self, s, start=None, finish=None, sources=None, with_timespan=False, with_source=False,
                 activity_journal=None, activity_group=None, bookmarks=None, warn_over=1):
        '''
        Specify any general constraints when constructing the object, then request particular statistics
        using by_name and by_group.

        The activity_group argument is used only to constrain any activity_journal (which may be given by date).

        The final dataframe can be retrieved directly via df or, via with_, additional processing can
        be made to rename columns, add statistics, etc.
        '''
        self.__s = s
        self.__start = start
        self.__finish = finish
        self.__sources = sources if sources else []
        if activity_journal:
            if not isinstance(activity_journal, Source):
                activity_journal = ActivityJournal.at(s, activity_journal, activity_group=activity_group)
            self.__sources.append(activity_journal)
        self.__with_timespan = with_timespan
        self.__with_source = with_source
        self.__activity_group = activity_group
        self.__warn_over = warn_over
        self.__statistic_names = {}
        self.__df = None
        if bookmarks: raise Exception('TODO')

    def __save_name(self, statistic_name):
        if statistic_name.name in self.__statistic_names:
            if statistic_name != self.__statistic_names[statistic_name.name]:
                raise Exception(f'Ambiguous name {statistic_name.name}')
        self.__statistic_names[statistic_name.name] = statistic_name

    def __name_and_type(self, name, owner, like):
        owner_name = short_cls(owner) if not isinstance(owner, str) else owner
        try:
            q = self.__s.query(StatisticName).filter(StatisticName.owner == owner)
            if like:
                statistic_names = q.filter(StatisticName.name.ilike(name)).all()
            else:
                statistic_names = [q.filter(StatisticName.name == name).one()]
            found = 0
            for statistic_name in statistic_names:
                self.__save_name(statistic_name)
                type_class = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]
                found += 1
                yield statistic_name, type_class
            if not found:
                log.warning(f'Did not match {owner_name}.{name} (like={like})')
        except Exception:
            log_current_exception(traceback=False)
            log.warning(f'Could not match {owner_name}.{name} (like={like})')

    def __columns(self, type_class, label):
        columns = [type_class.time.label(N.INDEX), type_class.value.label(label)]
        if self.__with_source:
            columns += [type_class.source_id.label(N._src(label))]
        return columns

    def by_name(self, owner, *names, like=False):
        for name in names:
            for statistic_name, type_class in self.__name_and_type(name, owner, like):
                label = statistic_name.name
                log.info(f'Retrieving {label}')
                q = self.__s.query(*self.__columns(type_class, label)). \
                    filter(type_class.statistic_name_id == statistic_name.id)
                q = self.__constrain_journal(q).order_by(N.INDEX)
                with timing(f'Slow query for {label}?\n{q}', self.__warn_over):
                    df = read_query(q, index=N.INDEX)
                self.__merge(df)
        return self

    def by_group(self, owner, *names, like=False):
        for name in names:
            for statistic_name, type_class in self.__name_and_type(name, owner, like):
                q_group_ids = self.__s.query(distinct(Source.activity_group_id)). \
                    join(StatisticJournal, StatisticJournal.source_id == Source.id). \
                    filter(StatisticJournal.statistic_name_id == statistic_name.id)
                q_group_ids = self.__constrain_journal(q_group_ids)
                with timing(f'Slow query for group ids\n{q_group_ids}?', self.__warn_over):
                    activity_group_ids = q_group_ids.all()
                for row in activity_group_ids:
                    activity_group_id = row[0]
                    if activity_group_id:
                        activity_group = self.__s.query(ActivityGroup). \
                            filter(ActivityGroup.id == activity_group_id).one()
                        label = statistic_name.name + ':' + activity_group.name
                    else:
                        label = statistic_name.name
                    log.info(f'Retrieving {label}')
                    q = self.__s.query(*self.__columns(type_class, label)). \
                        filter(type_class.statistic_name_id == statistic_name.id). \
                        join(Source, type_class.source_id == Source.id). \
                        filter(Source.activity_group_id == activity_group_id)
                    q = self.__constrain_journal(q).order_by(N.INDEX)
                    with timing(f'Slow query for {label}?\n{q}', self.__warn_over):
                        df = read_query(q, index=N.INDEX)
                    self.__merge(df)
        return self

    def __constrain_journal(self, q):
        if self.__start: q = q.filter(StatisticJournal.time >= self.__start)
        if self.__finish: q = q.filter(StatisticJournal.time < self.__finish)
        if self.__sources:
            q = q.filter(StatisticJournal.source_id.in_([source.id for source in self.__sources]))
        elif self.__activity_group:
            # only use of sources not specified separately (since those fix groups anyway)
            source = aliased(Source)
            q = q.join(source, source.id == StatisticJournal.source_id). \
                filter(source.activity_group_id == self.__activity_group.id)
        return q

    def __merge(self, df):
        if self.__df is None:
            self.__df = df
        else:
            with timing(f'Slow merge of {df.columns}?', self.__warn_over):
                self.__df = self.__df.join(df, how='outer')

    def __add_timespan(self):
        self.__df[N.TIMESPAN_ID] = np.nan
        for id, start, finish in self.__s.query(ActivityTimespan.id,
                                                ActivityTimespan.start, ActivityTimespan.finish). \
                filter(ActivityTimespan.activity_journal_id.in_([source.id for source in self.__sources])):
            self.__df.loc[start:finish, [N.TIMESPAN_ID]] = id

    @property
    def df(self):
        if self.__df is None:
            log.warning('Have no data!')
            self.__df = pd.DataFrame()  # if everything failed, allow code to continue with no data
        if self.__with_timespan:
            self.__add_timespan()
            self.__with_timespan = False
        return self.__df

    @property
    def with_(self):
        return Data(self.df, self.__statistic_names)


class Data:

    def __init__(self, df, statistic_names):
        self.df = df
        self.__statistic_names = statistic_names
        log.debug(f'Columns: {", ".join(self.df.columns)}')

    def __bool__(self):
        return not self.df.dropna(how='all').empty

    def drop_prefix(self, prefix, spaces=True):
        for column in like(prefix + '%', self.df.columns):
            new_column = column.replace(prefix, '')
            if spaces:
                while new_column.startswith(SPACE) or new_column.startswith(' '):
                    new_column = new_column[1:]
            log.debug(f'Rename {column} -> {new_column}')
            self.df.rename(columns={column: new_column}, inplace=True)
        return self
    
    def transform(self, name, scale=1.0, median=None):
        if scale != 1:
            self.df[name] = self.df[name] * scale
        if median:
            self.df[name] = self.df[name].rolling(median, min_periods=1).median()

    def __rename(self, name, new_name, scale=1.0, median=None):
        log.debug(f'{name} -> {new_name}')
        self.df.rename(columns={name: new_name}, inplace=True)
        self.transform(new_name, scale=scale, median=median)

    def __copy(self, name, new_name, scale=1.0, median=None):
        log.debug(f'{name} <-> {new_name}')
        self.df[new_name] = self.df[name]
        self.transform(new_name, scale=scale, median=median)

    def __with_names_values(self, op, map, scale=1.0, median=None):
        for name, value in map.items():
            if present(self.df, name):
                op(name, value, scale=scale, median=median)
            else:
                log.warning(f'Missing {name} in data')
        return self

    def rename(self, map, scale=1.0, median=None):
        return self.__with_names_values(self.__rename, map, scale=scale, median=median)

    def copy(self, map, scale=1.0, median=None):
        return self.__with_names_values(self.__copy, map, scale=scale, median=median)

    def __with_names_units(self, op, columns):
        for column in columns:
            if ':' in column:
                name, group = column.split(':', 1)
            else:
                name, group = column, None
            if name in self.__statistic_names:   # may not have any data
                statistic_name = self.__statistic_names[name]
                if group:
                    new_name = N._slash(name, statistic_name.units) + ':' + group
                else:
                    new_name = N._slash(name, statistic_name.units)
                op(column, new_name)
        return self

    def rename_with_units(self, *columns):
        if not columns: columns = self.df.columns
        return self.__with_names_units(self.__rename, columns)

    def copy_with_units(self, *columns):
        if not columns: columns = self.df.columns
        return self.__with_names_units(self.__copy, columns)

    def into(self, df, tolerance, interpolate=False):
        if self:
            extra = self.df
            if interpolate: extra = extra.interpolate(method='time')
            extra = extra.reindex(df.index, method='nearest', tolerance=tolerance)
            return df.merge(extra, how='left', left_index=True, right_index=True)
        else:
            return df

    def add_times(self):
        set_times_from_index(self.df)
        return self

    def __coallesce(self, columns, name, delete=False):
        log.debug(f'Coallescing {columns} for {name}')
        df = self.df[columns].copy()
        df.fillna(method='ffill', axis='columns', inplace=True)
        df.fillna(method='bfill', axis='columns', inplace=True)
        self.df.loc[:, name] = df.iloc[:, [0]]
        if delete:
            self.df.drop(columns=columns, inplace=True)

    def coallesce_groups(self, *names, delete=False):
        '''
        Merge columns named with groups.
        '''
        if not names:
            names = set(column.split(':')[0] for column in self.df.columns if ':' in column)
        log.debug(f'Coallescing groups for {names}')
        for name in names:
            if name in self.df.columns:
                raise Exception(f'{name} already exists')
            columns = like(name + ':%', self.df.columns)
            self.__coallesce(columns, name, delete=delete)
        return self

    def coallesce_like(self, map, delete=False):
        '''
        Merge columns that match.  map is {pattern: result}
        '''
        log.debug(f'Coallescing {map}')
        for pattern in map:
            name = map[pattern]
            if name in self.df.columns:
                raise Exception(f'{name} already exists')
            columns = like(pattern, self.df.columns)
            self.__coallesce(columns, name, delete=delete)
        return self


def set_times_from_index(df):
    df.loc[:, N.TIME] = pd.to_datetime(df.index)
    df.loc[:, N.LOCAL_TIME] = df[N.TIME].apply(lambda x: time_to_local_time(x.to_pydatetime(), YMD))


def std_health_statistics(s, freq='1h'):

    from ..pipeline.owners import RestHRCalculator, ResponseCalculator, ActivityCalculator, MonitorCalculator

    start = s.query(StatisticJournal.time). \
        filter(StatisticJournal.time > local_date_to_time(to_date('1970-01-03'))). \
        order_by(asc(StatisticJournal.time)).limit(1).scalar()
    finish = s.query(StatisticJournal.time).order_by(desc(StatisticJournal.time)).limit(1).scalar()

    stats = pd.DataFrame(index=pd.date_range(start=start, end=finish, freq=freq))
    set_times_from_index(stats)

    stats = Statistics(s). \
        by_name(ResponseCalculator, N.DEFAULT_ANY, like=True).with_. \
        drop_prefix(N.DEFAULT + SPACE).into(stats, tolerance='30m')

    stats = Statistics(s). \
        by_name(RestHRCalculator, N.REST_HR). \
        by_name(MonitorCalculator, N.DAILY_STEPS). \
        by_name(ActivityCalculator, N._delta(N.DEFAULT_ANY), like=True).with_. \
        rename_with_units(N.REST_HR).into(stats, tolerance='30m')

    stats = Statistics(s).\
        by_group(ActivityCalculator, N.ACTIVE_TIME, N.ACTIVE_DISTANCE).with_. \
        coallesce_groups(N.ACTIVE_TIME, N.ACTIVE_DISTANCE). \
        rename_with_units(N.ACTIVE_TIME, N.ACTIVE_DISTANCE). \
        copy({N.ACTIVE_TIME_S: N.ACTIVE_TIME_H}, scale=1 / 3600). \
        into(stats, tolerance='30m')

    return stats


def std_activity_statistics(s, activity_journal, activity_group=None):

    # the choice of which values have units is somewhat arbitrary, but less so than it was...

    from ..pipeline.calculate.elevation import ElevationCalculator
    from ..pipeline.calculate.impulse import ImpulseCalculator
    from ..pipeline.calculate.power import PowerCalculator
    from ..pipeline.read.segment import SegmentReader

    if not isinstance(activity_journal, ActivityJournal):
        activity_journal = ActivityJournal.at(s, activity_journal, activity_group=activity_group)

    stats = Statistics(s, activity_journal=activity_journal, with_timespan=True). \
        by_name(SegmentReader, N.LATITUDE, N.LONGITUDE, N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y,
                N.DISTANCE, N.SPEED, N.CADENCE, N.ALTITUDE, N.HEART_RATE).with_. \
        rename_with_units(N.LATITUDE, N.LONGITUDE, N.DISTANCE, N.SPEED, N.CADENCE, N.ALTITUDE, N.HEART_RATE). \
        copy({N.SPEED_MS: N.MED_SPEED_KMH}, scale=3.6, median=MED_WINDOW). \
        copy({N.HEART_RATE_BPM: N.MED_HEART_RATE_BPM}, median=MED_WINDOW). \
        copy({N.CADENCE_RPM: N.MED_CADENCE_RPM}, median=MED_WINDOW). \
        add_times().df

    stats = Statistics(s, activity_journal=activity_journal). \
        by_name(ElevationCalculator, N.ELEVATION, N.GRADE).with_. \
        rename_with_units().into(stats, tolerance='1s')

    hr_impulse_10 = N.DEFAULT + SPACE + N.HR_IMPULSE_10
    stats = Statistics(s, activity_journal=activity_journal). \
        by_name(ImpulseCalculator, N.HR_ZONE, hr_impulse_10).with_. \
        drop_prefix(N.DEFAULT + SPACE).into(stats, tolerance='10s', interpolate=True)

    stats = Statistics(s, activity_journal=activity_journal). \
        by_name(PowerCalculator, N.POWER_ESTIMATE).with_. \
        rename_with_units(). \
        copy({N.POWER_ESTIMATE_W: N.MED_POWER_ESTIMATE_W}, median=MED_WINDOW). \
        into(stats, tolerance='1s')

    return stats


if __name__ == '__main__':

    from ..pipeline.owners import ActivityCalculator, SegmentReader, MonitorCalculator

    s = session('-v5')

    with timing('select'):
        # df = std_health_statistics(s)
        # df = std_activity_statistics(s, '2020-05-15', 'road')
        # df = Statistics(s).like(N.CLIMB_ANY, owner=ActivityCalculator).from_(activity_journal='2020-05-15').by_group().df
        # df = Statistics(s).for_(N.ACTIVE_DISTANCE, owner=ActivityCalculator).by_group()
        # acc = Accumulator(s, sources=[ActivityJournal.at(s, '2020-05-15')])
        # acc = Accumulator(s, with_timespan=True, sources=[ActivityJournal.at(s, '2020-05-15')])
        # acc = Accumulator(s, with_source=True)
        # acc.by_name(ActivityCalculator, N.CLIMB_ANY, like=True)
        # df = acc.df
        df = Statistics(s). \
            by_name(ActivityTopic, N.NAME). \
            by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB).with_. \
            copy({N.ACTIVE_DISTANCE: N.ACTIVE_DISTANCE_KM}).add_times().df
        print(df.describe())
        df['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)

    print(df)
    print(df.describe())
    print(df.columns)
    print(df[N.REST_HR_BPM].describe())




from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np
import pandas as pd

from .utils import ActivityGroupCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator
from ...data import present, linear_resample_time, Statistics
from ...data.frame import median_dt
from ...data.lib import interpolate_to_index
from ...data.power import add_differentials, add_energy_budget, add_loss_estimate, add_power_estimate, PowerException, \
    add_air_speed
from ...lib import log_current_exception
from ...lib.data import reftuple, MissingReference
from ...names import N, Units, Summaries, T, simple_name
from ...sql import StatisticJournalFloat, Constant

log = getLogger(__name__)

BikeModel = namedtuple('BikeModel', 'cda, crr, bike_weight')


class PowerModel(reftuple('Power', 'bike_model, rider_weight')):

    def expand(self, s, time, default_owner=None, default_activity_group=None):
        instance = super().expand(s, time, default_owner=default_owner, default_activity_group=default_activity_group)
        bike_model = instance.bike_model
        if not isinstance(bike_model, BikeModel):
            return instance._replace(bike_model=BikeModel(**bike_model))
        else:
            return instance


class PowerCalculator(ActivityGroupCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    '''
    See ch2.config.power for examples of how this is configured.
    '''

    def __init__(self, *args, power_model=None, caloric_eff=0.25, **kargs):
        self.power_model_ref = power_model
        self.caloric_eff = caloric_eff
        super().__init__(*args, **kargs)

    def _set_power(self, s, ajournal):
        power_model = PowerModel(**loads(Constant.from_name(s, self.power_model_ref).at(s).value))
        self.power_model = power_model.expand(s, ajournal.start,
                                              default_owner=Constant, default_activity_group=self.activity_group)
        log.debug(f'Power: {self.power_model_ref}: {self.power_model}')

    def _read_dataframe(self, s, ajournal):
        from ..owners import SegmentReader, ElevationCalculator
        try:
            self._set_power(s, ajournal)
            df = Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(SegmentReader, N.DISTANCE, N.SPEED, N.CADENCE, N.LATITUDE,
                        N.LONGITUDE, N.HEART_RATE). \
                by_name(ElevationCalculator, N.ELEVATION).df
            ldf = linear_resample_time(df)
            ldf = add_differentials(ldf, max_gap=1.1 * median_dt(df))
            if N.HEADING not in ldf.columns:
                raise PowerException('Could not calculate heading')    
            return df, ldf
        except PowerException as e:
            log.warning(e)
        except MissingReference as e:
            log.warning(f'Power configuration incorrect ({e})')
        except Exception as e:
            log.warning(f'Failed to generate statistics for power ({ajournal.activity_group.name}): {e}')
            log_current_exception()

    def _calculate_stats(self, s, ajournal, dfs):
        df, ldf = dfs
        total_weight = self.power_model.bike_model.bike_weight + self.power_model.rider_weight
        ldf = add_energy_budget(ldf, total_weight)
        ldf = add_air_speed(ldf, 0, 0)
        ldf = add_loss_estimate(ldf, total_weight, cda=self.power_model.bike_model.cda,
                                crr=self.power_model.bike_model.crr)
        ldf = add_power_estimate(ldf)
        return df, ldf

    def _copy_results(self, s, ajournal, loader, dfs,
                      fields=((T.POWER_ESTIMATE, Units.W, Summaries.AVG, 'The estimated power.'),
                              (T.HEADING, Units.DEG, None, 'The current heading'))):
        df, ldf = dfs
        self.__add_total_energy(s, ajournal, loader, ldf)
        df = interpolate_to_index(df, ldf, *(simple_name(field[0]) for field in fields))
        for time, row in df.iterrows():
            for title, units, summary, description in fields:
                name = simple_name(title)
                if name in row and not pd.isnull(row[name]):
                    loader.add(name, units, summary, ajournal, row[name], time,
                               StatisticJournalFloat, title=title, description=description)

    def __add_total_energy(self, s, ajournal, loader, ldf):
        if present(ldf, N.POWER_ESTIMATE):
            ldf['tmp'] = ldf[N.POWER_ESTIMATE]
            ldf.loc[ldf['tmp'].isna(), ['tmp']] = 0
            energy = np.trapz(y=ldf['tmp'], x=ldf.index.astype(np.int64) / 1e12)
            loader.add(T.ENERGY_ESTIMATE, Units.KJ, Summaries.MAX, ajournal, energy, ajournal.start,
                       StatisticJournalFloat, 'The estimated total energy expended.')
            loader.add(T.CALORIE_ESTIMATE, Units.KCAL, Summaries.MAX, ajournal,
                       energy * 0.239006 / self.caloric_eff, ajournal.start, StatisticJournalFloat,
                       'The estimated calories burnt.')
            ldf.drop(columns=['tmp'], inplace=True)
